# coding: utf8
import clinica.utils.filemanip
import clinica.utils.participant

__author__ = "Alexis Guyot"
__copyright__ = "Copyright 2016-2019, The Aramis Lab Team"
__credits__ = ["Alexis Guyot"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Alexis Guyot"
__email__ = "alexis.guyot@icm-institute.org"
__status__ = "Development"


def get_capsxsectional_path(caps_dir, subject, session):
    """Get path to cross-sectional run in CAPS dir

    Args:
        caps_dir (string): location of CAPS folder
        subject (string): participant ID
        session_list (strings): session

    Returns:
        caps_xsectional_path (string): path to the cross-sectional
            subfolder in the CAPS directory
    """
    import os

    caps_xsectional_path = os.path.join(
        caps_dir,
        'subjects',
        subject,
        session,
        't1',
        'freesurfer_cross_sectional'
        )

    return caps_xsectional_path


def check_xsectional_reconalled(caps_dir, subject_list, session_list):
    """Check cross-sectional recon-all was run previously

    Checks if t1-freesurfer-cross-sectional has been
    correctly run beforehand.

    Args:
        caps_dir (string): CAPS directory to contain the output
        subject_list (list): a list containing all the participant_id
        session_list (list): a list containing all the session_id

    Returns:
        N/A
    """
    import os
    import subprocess
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils

    # subses_list, subject_list and session_list
    subject_number = len(subject_list)
    subses_list = list(
        '{0}_{1}'.format(
            subject_list[subject_index],
            session_list[subject_index]
            ) for subject_index in range(subject_number))

    # change the relative path to be absolute path
    abs_caps_path = os.path.abspath(caps_dir)
    subjects_path = os.path.join(abs_caps_path, 'subjects')

    if not os.path.isdir(subjects_path):
        error_msg = 'Error: {0} does not exist.'.format(subjects_path)
        raise IOError(error_msg)

    for subject_index in range(subject_number):
        subject = subject_list[subject_index]
        session = session_list[subject_index]
        subses = subses_list[subject_index]

        xsectional_path = utils.get_capsxsectional_path(
            caps_dir, subject, session)
        if not os.path.isdir(xsectional_path):
            error_msg = 'Error: {0} does not exist.'.format(xsectional_path)
            raise IOError(error_msg)

        # check the recon-all.log
        log_file = os.path.join(
            xsectional_path, subses, 'scripts', 'recon-all.log')
        if os.path.isfile(log_file):
            last_line = subprocess.check_output(['tail', '-1', log_file])
            if b'finished without error' not in last_line:
                error_msg = 'Subject {0} has not been processed yet.'.format(
                    subses)
                error_msg = '{0} Re-run t1-freesurfer-cross-sectional.'.format(
                    error_msg)
                raise IOError(error_msg)


def sessionid_to_sessionlabel(sessionid):
    """Extract session label from session ID

    Example:
    if session_id='ses_M00', the session_label='M00'

    Args:
        sessionid (string): session ID

    Returns:
        sessionlabel (string): session ID
    """
    sessionlabel = sessionid[4:]

    return sessionlabel


def get_longsubdir_name(session_list):
    """Returns a subfolder name associated to a set of visits

    This will create a unique identifier for a subject and its
    corresponding sessions.
    Used to be able to run the longitudinal processing at different
    points in time, where new visits are successively added.
    eg.
    - 1. t=0 year, subject: sub01, sessions: ses-M00, ses-M01
        -> identifier: long-M00M01
    - 2. t=1 year, subject: sub01, sessions: ses-M00, ses-M01, ses-M02
        -> identifier: long-M00M01M02
    - 2. t=2 year, subject: sub01, sessions: ses-M00, ses-M01, ses-M02,
            ses-M03, ses-M04
        -> identifier: long-M00M01M02M03M04

    Args:
        session_list (list of strings):

    Returns:
        long_subdirname (string): name of the subfolder
            corresponding to all the sessions of a unique subject. This
            will be used as part of both:
            - the path to the unbiased template for a subject and its
                sessions
            - the corresponding path for the longitudinal correction
    """
    # get session label for each session ID
    sessionlabel_list = [
        sessionid_to_sessionlabel(session) for session in session_list]

    # concatenate all the sessions labels
    long_subdirname = ''.join(sessionlabel_list)
    long_subdirname = 'long-{0}'.format(long_subdirname)

    return long_subdirname


def get_capstemplate_path(caps_dir, subject, session_list):
    """Get path to template in CAPS dir

    Args:
        caps_dir (string): location of CAPS folder
        subject (string): participant ID
        session_list (list of strings): list of all the sessions
            corresponding to the participant

    Returns:
        caps_template_path (string): path to the template subfolder in the
            CAPS directory
    """
    import os
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils

    # get unique identifier corresponding to the subject and all
    # associated sessions. Will be used as part of a path in the CAPS
    # dir
    long_subdirname = utils.get_longsubdir_name(session_list)
    # define template location in the CAPS directory
    caps_template_path = os.path.join(
        caps_dir,
        'subjects',
        subject,
        long_subdirname,
        'freesurfer_unbiased_template',
        subject)

    return caps_template_path


def get_longsubdir_dict(
        in_subject_list, in_session_list):
    """Get subdir name for each subject

    For each unique subject in the provided subject list, return the
    name of the longitudinal subfolder that is associated to the
    current set of visits

    Args:
        in_subject_list (list of string): list of participant_id
        in_session_list (list of string): list of sessions

    Returns:
        longsubdir_dict (dictionary of strings): for each
            unique subject, name of longitudinal subofolder associated
            to the current set of visits
    """
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils

    # get list of unique subjects and for each individual subjects get
    # list of associated sessions
    [
        unique_subject_list,
        persubject_session_list2] = clinica.utils.participant.get_unique_subjects(
            in_subject_list, in_session_list)

    # get longitudinal identifier for ech subject and store into
    # dictionary
    longsubdir_dict = dict()
    subject_number = len(unique_subject_list)
    for subject_index in range(subject_number):
        subject = unique_subject_list[subject_index]
        subject_session_list = persubject_session_list2[subject_index]
        longsubdir_dict[subject] = utils.get_longsubdir_name(
            subject_session_list)

    return longsubdir_dict


def get_capslongcor_path(caps_dir, subject, session, longsubdir_dict):
    """Get path to longitudinal correction subfolder in CAPS dir

    Args:
        caps_dir (string): location of CAPS folder
        subject (string): participant ID
        session (string): session ID
        longsubdir_dict (dictionary of strings): for each
            unique subject, name of longitudinal subofolder associated
            to the current set of visits

    Returns:
        caps_longcor_path (string): path to the longitudinal correction
            subfolder in the CAPS directory
    """
    import os

    # get longitudinal subfolder identifier for the current subject ID
    longsubdir = longsubdir_dict[subject]

    # get target CAPS subfolder
    capslongcor_path = os.path.join(
        caps_dir,
        'subjects',
        subject,
        session,
        't1',
        longsubdir,
        'freesurfer_longitudinal',
        '{0}_{1}.long.{0}'.format(subject, session))

    return capslongcor_path


def boolean_overwrite_caps(str_overwrite_caps):
    """Convert overwrite-caps flag to Boolean

    Args:
        str_overwrite_caps (string): any option in 'true', 'True',
            'TRUE', 'false', 'False', 'FALSE'

    Returns:
        overwrite_caps (Boolean): corresponding boolean value
    """
    if str_overwrite_caps in ['true', 'True', 'TRUE']:
        overwrite_caps = True
    elif str_overwrite_caps in ['false', 'False', 'FALSE']:
        overwrite_caps = False
    else:
        error_msg = 'Error: --force-overwrite should be \'true\' or \'false\''
        raise ValueError(error_msg)

    return overwrite_caps


def get_cl_overwritecaps(
        caps_dir,
        tsv_path,
        working_directory,
        n_procs):
    """Define command-line to force overwriting CAPS folder

    Output the command line a user will need to run if they want to
    force overwriting the CAPS folder, using the same parameters as
    were passed to the terminal for the current execution of the
    pipeline.

    Args:
        caps_dir (string): location of CAPS folder
        tsv_path (string): path to .tsv file provided by user
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
            None if no subjects already processed.
        working_directory (string): path to working directory
            provided by user. None if they did not provide any.
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        n_procs (int): number of processors provided by user
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.

    Returns:
        cl_overwrite_caps (string): command line to force overwriting
            the CAPS folder. Parameters (e.g. CAPS dir, tsv file,
            working directory and n_procs) are the same as those
            provided by the user for the current execution of the
            pipeline.
    """
    # initialise
    cl_overwrite_caps = 'clinica run t1-freesurfer-longitudinal'
    # CAPS dir: mandatory argument
    cl_overwrite_caps = '{0} {1}'.format(cl_overwrite_caps, caps_dir)
    # TSV file: optional
    if tsv_path is not None:
        cl_overwrite_caps = '{0} -tsv {1}'.format(cl_overwrite_caps, tsv_path)
    # working directory: optional
    if working_directory is not None:
        cl_overwrite_caps = '{0} -wd {1}'.format(
            cl_overwrite_caps, working_directory)
    # number of processors: optional
    if n_procs is not None:
        cl_overwrite_caps = '{0} -np {1}'.format(cl_overwrite_caps, n_procs)
    # option to force overwrite: mandatory
    cl_overwrite_caps = '{0} --overwrite_caps True'.format(cl_overwrite_caps)

    return cl_overwrite_caps


def check_caps_processing(
        processing_type,
        subject_list,
        sesobj_list,
        caps_dir):
    """Check there is no processing done in CAPS already

    Check, for any of the 'template' or 'longitudinal-correction'
    processing type which subjects/sessions have been / have not been processed already.
    correspond folder for the current subject and its corresponding
    sessions.
    Three cases are possible:
    1. CAPS folder contains no processing
    2. All subjects/sessions have already been processed
    3. Some subjects/sessions (but not all) have been processed already.

    Args:
        processing_type (string): Either 'template' or 'longitudinal
            correction'. Used to build a warning message specific to the
            type of data that is being looked for in the CAPS folder.
        subject_list (list of string): list of participant_id. Will
            either be a list of unique subjects (template) or a list of
            all subjects for all sub-sess (longitudinal-correction)
        sesobj_list (list of objects): either a list of list of sessions
            for all unique participant_id (template) or a list of
            sessions (longitudinal_correction)
        caps_dir (string): location of CAPS folder

    Returns:
        pcssd_sublist (list of strings): list of all the subjects for
            which processing was found in the CAPS folder
        pcssd_sesobjlist (list of objects): list, for all subjects
            for which processing was found in the CAPS folder, of either
            the list of associated session (template) or of single
            sessions (longitudinal correction)
        unpcssd_capstargetlist (string): list, for all subjects for
            which processing was found in the CAPS folder, of
            associated CAPS folder where the template or
            longitudinal-correction processing will get stored
        unpcssd_sublist (list of strings): list of all the subjects
            for which no processing was found in the CAPS folder
        unpcssd_sesobjlist (list of objects): list, for all subjects
            for which no processing was found in the CAPS folder, of
            either the list of associated session (template) or of
            single sessions (longitudinal correction)
        unpcssd_capstargetlist (string): list, for all subjects for
            which no processing was found in the CAPS folder, of
            associated CAPS folder where the template or
            longitudinal-correction processing will get stored
        all_sublist (list of strings): list of all the subjects
        all_sesobjlist (list of objects): list, for all subjects of
            either the list of associated session (template) or of
            single sessions (longitudinal correction)
        all_capstargetlist (string): list, for all subjects and
            corresponding sessions of associated CAPS folder where the
            template will get stored
    """
    import os
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils

    # initialise lists of CAPS targets and list of processed /
    # unprocessed subjects
    pcssd_sublist = list()
    pcssd_sesobjlist = list()
    pcssd_capstargetlist = list()
    unpcssd_sublist = list()
    unpcssd_sesobjlist = list()
    unpcssd_capstargetlist = list()
    all_sublist = list()
    all_sesobjlist = list()
    all_capstargetlist = list()

    # if dealing with longitudinal-correction, get longitudinal
    # subfolder identifier for each unique subject prior to looping over
    # each sub/ses
    if processing_type == 'longitudinal-correction':
        longsubdir_dict = get_longsubdir_dict(subject_list, sesobj_list)

    # loop through all subjects and their associated sesobj
    subject_number = len(subject_list)
    for subject_index in range(subject_number):
        subject = subject_list[subject_index]
        sesobj = sesobj_list[subject_index]
        # get location in the CAPS directory
        if processing_type == 'template':
            caps_target = utils.get_capstemplate_path(
                caps_dir, subject, sesobj)
        elif processing_type == 'longitudinal-correction':
            caps_target = utils.get_capslongcor_path(
                caps_dir, subject, sesobj, longsubdir_dict)
        all_sublist.append(subject)
        all_sesobjlist.append(sesobj)
        all_capstargetlist.append(caps_target)
        # check if there is already a template in the CAPS directory
        if os.path.isdir(caps_target):
            # there is a template already. Update list of processed
            # subjects
            pcssd_sublist.append(subject)
            pcssd_sesobjlist.append(sesobj)
            pcssd_capstargetlist.append(caps_target)
        else:
            # there is no template folder in CAPS corresponding to the
            # subject. Update list of unprocessed subjects and
            # unprocessed sessions
            unpcssd_sublist.append(subject)
            unpcssd_sesobjlist.append(sesobj)
            unpcssd_capstargetlist.append(caps_target)

    return [
        pcssd_sublist,
        pcssd_sesobjlist,
        pcssd_capstargetlist,
        unpcssd_sublist,
        unpcssd_sesobjlist,
        unpcssd_capstargetlist,
        all_sublist,
        all_sesobjlist,
        all_capstargetlist]


def check_caps_template(
        subject_list,
        session_list,
        caps_dir):
    """Check there is not a template in CAPS already

    Check the CAPS folder does not already contain a template folder
    for the current subject and its corresponding sessions.
    Three cases are possible:
    1. CAPS folder contains no processed templates
    -> return target path for all subjects/sessions
    2. All templates have already been processed and sent to CAPS folder
    -> if force_overwrite option provided, return target path for all
        subjects/sessions, else exit
    3. Some templates (but not all) have been processed already.
    -> if force_overwrite option provided, return target path for all
        subjects/sessions, else return target path for the
        subjects/sessions that have not been processed yet

    Args:
        subject_list (list of string): list of participant_id
        session_list (list of string): list of session_id corresponding
            to a specific participant_id
        caps_dir (string): location of CAPS folder

    Returns:
        pcssd_sublist (list of strings): list of all the subjects for
            which processing was found in the CAPS folder
        pcssd_seslist2 (list of objects): list, for all subjects
            for which processing was found in the CAPS folder, of the
            list of associated sessions
        unpcssd_sublist (list of strings): list of all the subjects
            for which no processing was found in the CAPS folder
        unpcssd_seslist2 (list of objects): list, for all subjects
            for which no processing was found in the CAPS folder, of the
            list of associated session
        unpcssd_capstargetlist (string): list, for all subjects for
            which no processing was found in the CAPS folder, of
            associated CAPS folder where the template or
            longitudinal-correction processing will get stored
        all_sublist (list of strings): list of all the subjects
        all_seslist2 (list of objects): list, for all subjects of
            the list of associated sessions
        all_capstargetlist (string): list, for all subjects and
            corresponding sessions of associated CAPS folder where the
            template will get stored
    """
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils

    # get a unique list of subjects and associated per-subject list of
    # sessions
    unique_subject_list, persubject_session_list2 = clinica.utils.participant.get_unique_subjects(
        subject_list, session_list)

    # Check processed/unprocessed subjects
    [
        pcssd_sublist,
        pcssd_seslist2,
        pcssd_capstargetlist,
        unpcssd_sublist,
        unpcssd_seslist2,
        unpcssd_capstargetlist,
        all_sublist,
        all_seslist2,
        all_capstargetlist] = utils.check_caps_processing(
            'template',
            unique_subject_list,
            persubject_session_list2,
            caps_dir)

    return [
        pcssd_sublist,
        pcssd_seslist2,
        pcssd_capstargetlist,
        unpcssd_sublist,
        unpcssd_seslist2,
        unpcssd_capstargetlist,
        all_sublist,
        all_seslist2,
        all_capstargetlist]


def to_process(
        pcssd_sublist,
        pcssd_seslist2,
        unpcssd_sublist,
        unpcssd_seslist2,
        unpcssd_capstargetlist,
        all_sublist,
        all_seslist2,
        all_capstargetlist,
        in_caps_dir,
        in_overwrite_caps,
        in_working_directory,
        in_n_procs,
        in_rundir):
    """ Get list of subjects to be processed

    The list of subjects to be processed will depend upon whether the
    user provided the option to --force-overwrite already existing
    processing or not. If the option --force-overwrite was not passed
    but existing processing was found, a .tsv file will be generated
    that contains the list of subjects/sessions already processed.

    Args:
        pcssd_sublist (list of strings): list of all the subjects for
            which processing was found in the CAPS folder
        pcssd_seslist2 (list of objects): list, for all subjects
            for which processing was found in the CAPS folder, of the
            list of associated sessions
        unpcssd_sublist (list of strings): list of all the subjects
            for which no processing was found in the CAPS folder
        unpcssd_seslist2 (list of objects): list, for all subjects
            for which no processing was found in the CAPS folder, of the
            list of associated session
        unpcssd_capstargetlist (string): list, for all subjects for
            which no processing was found in the CAPS folder, of
            associated CAPS folder where the template or
            longitudinal-correction processing will get stored
        all_sublist (list of strings): list of all the subjects
        all_seslist2 (list of objects): list, for all subjects of
            the list of associated sessions
        all_capstargetlist (string): list, for all subjects and
            corresponding sessions of associated CAPS folder where the
            template will get stored
        in_caps_dir (string): CAPS directory to contain the output
        in_overwrite_caps (string): Option provided by the user to state
            whether the data in the CAPS folder (i.e.,
            previously-computed template) should be overwritten or not
            by a new run. By default, do not overwrite. Will overwrite
            for the following values: 'true', 'True' and 'TRUE'
        in_working_directory (string): path to working directory
            provided by user. None if they did not provide any.
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        in_n_procs (int): number of processors provided by user
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        in_rundir (string): path to the directory where clinica is being
            launched. Used to create a .tsv file in running directory
            if the user wishes to re-run the pipeline on subjects that
            had already been processed.

    Returns:
        topcss_sublist (list of strings): list of the subjects that will
            be processed. All if --force-overwrite, else only subjects
            that have not been processed already
        topcss_seslist2 (list of list): list of the list of sessions
            that correspond to each subject to be processed
        topcss_capstargetlist (list of strings): list, for all subjects
            to be processed, of the associated CAPS folder where the
            template will get stored
        overwrite_tsv_path (string): path to the .tsv to use if the user
            wishes to overwrite the subject that have already been
            processed in the CAPS folder. None if no subjects already
            processed.
    """
    import os
    import csv
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils

    # Convert overwrite-caps flag to boolean
    force_overwrite = utils.boolean_overwrite_caps(in_overwrite_caps)

    # Generate .tsv file. Will be used to re-run with --force-overwrite
    # enabled.
    pcssd_sub_number = len(pcssd_sublist)
    if pcssd_sub_number > 0:
        # only write the .tsv file if at least one subject already
        # processed
        overwrite_tsv_path = os.path.join(in_rundir, 'overwrite.tsv')
        with open(overwrite_tsv_path, 'w') as overwrite_tsv_file:
            # create tsv file
            tsv_writer = csv.writer(
                overwrite_tsv_file, delimiter='\t')
            tsv_writer.writerow(['participant_id', 'session_id'])
            # loop through all processed subject/session
            for pcssd_sub_index in range(pcssd_sub_number):
                pcssd_sub = pcssd_sublist[pcssd_sub_index]
                pcssd_seslist = pcssd_seslist2[pcssd_sub_index]
                pcssd_seslist_number = len(pcssd_seslist)
                for pcssd_seslist_index in range(pcssd_seslist_number):
                    pcssd_ses = pcssd_seslist[pcssd_seslist_index]
                    tsv_writer.writerow([pcssd_sub, pcssd_ses])
    else:
        overwrite_tsv_path = None

    # if all subjects already processed and --force-overwrite option not
    # provided, exit
    if (not unpcssd_sublist) and (not force_overwrite):
        cl_overwrite_caps = utils.get_cl_overwritecaps(
            in_caps_dir,
            overwrite_tsv_path,
            in_working_directory,
            in_n_procs)
        error_msg = 'Error: all subjects are already processed. To overwrite'
        error_msg = '{0} the data in CAPS folder, please run:\n'.format(
            error_msg)
        error_msg = '{0} {1}'.format(error_msg, cl_overwrite_caps)
        raise ValueError(error_msg)

    # check what subjects / sessions / CAPS target dirs to process
    # depending on --force-overwrite flag
    if force_overwrite:
        topcss_sublist = all_sublist
        topcss_seslist2 = all_seslist2
        topcss_capstargetlist = all_capstargetlist
    else:
        topcss_sublist = unpcssd_sublist
        topcss_seslist2 = unpcssd_seslist2
        topcss_capstargetlist = unpcssd_capstargetlist

    return [
        topcss_sublist,
        topcss_seslist2,
        topcss_capstargetlist,
        overwrite_tsv_path]


def check_single_timepoint(subject_list, session_list2):
    """Check and warn in case subject with 1 timepoint detected

    Args:
        sublist (list of strings): list of the subjects that will be
            processed. All if --force-overwrite, else only subjects that
            have not been processed already
        seslist2 (list of list): list of the list of sessions that
            correspond to each subject to be processed

    Returns:
        N/A
    """
    import sys
    import os
    import warnings
    import colorama

    # Loop through all subjects and associated session lists
    single_timepoint_detected = False
    single_timepoint_subject_list = list()
    for subject, subses_list in zip(subject_list, session_list2):
        if len(subses_list) == 1:
            single_timepoint_detected = True
            single_timepoint_subject_list.append(subject)

    # Warn the user if a subject was found with a single time point
    if single_timepoint_detected:
        # Retrieve the system temporary folder location
        # Check operating system
        platform = sys.platform
        if platform.startswith('linux'):
            # if Linux: location is /tmp
            temp_folder_location = ' (/tmp)'
        elif platform == 'darwin':
            # if OS X: location is given by $TMPDIR
            temp_folder_location = ' ({0})'.format(os.environ['TMPDIR'])
        else:
            # other OS (e.g., other Unix, cygwin). Not supported by Clinica
            temp_folder_location = ''
        # Build the warning message
        warning_msg = '{0}'.format(colorama.Fore.RED)
        if len(single_timepoint_subject_list) == 1:
            warning_msg = '{0}Only one session was detected for the following subject:'.format(
                warning_msg)
        else:
            warning_msg = '{0}Only one session was detected for the following subjects:'.format(
                warning_msg)
        for single_timepoint_subject in single_timepoint_subject_list:
            warning_msg = '{0}\n    - {1}'.format(warning_msg, single_timepoint_subject)
        if len(single_timepoint_subject_list) == 1:
            warning_msg = '{0}\n Processing for this subject'.format(warning_msg)
        else:
            warning_msg = '{0}\n Processing for these subjects'.format(warning_msg)
        warning_msg = '{0} will be done inside the system temporary folder'.format(warning_msg)
        warning_msg = '{0}{1}.'.format(warning_msg, temp_folder_location)
        warning_msg = '{0}{1}'.format(warning_msg, colorama.Fore.RESET)
        # Show the warning
        warnings.warn(warning_msg)


def process_input_node(
        in_caps_dir,
        in_subject_list,
        in_session_list,
        in_overwrite_caps,
        in_working_directory,
        in_n_procs,
        in_rundir):
    """Carry out all processing for the input node

    Performs all the processing for the input node:
    1. Check if t1-freesurfer-cross-sectional has been correctly run
        beforehand
    2. Check if there are already template subfolders in the CAPS dir.

    Args:
        in_caps_dir (string): CAPS directory to contain the output
        in_subject_list (list): a list containing all the participant_id
        in_session_list (list): a list containing all the session_id
        in_overwrite_caps (string): Option provided by the user to state
            whether the data in the CAPS folder (i.e.,
            previously-computed template) should be overwritten or not
            by a new run. By default, do not overwrite. Will overwrite
            for the following values: 'true', 'True' and 'TRUE'
        in_working_directory (string): path to working directory
            provided by user. None if they did not provide any.
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        in_n_procs (int): number of processors provided by user
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        in_rundir (str): path to the directory where clinica is being
            launched. Used to create a .tsv file in running directory
            if the user wishes to re-run the pipeline on subjects that
            had already been processed.

    Returns:
        out_subject_list (list of string): list of participant_id, where
            each participant appears only once
        out_session_list2 (list of list): list of list (list2) of
            session_id associated to any single participant
        out_capstarget_list (list of string): list of path to the
            template directory in CAPS folder for each participant
        out_unpcssd_sublist (list of string): list of participant_id,
            where each participant appears only once, wich have not
            already been (template) processed
        out_pcssd_capstargetlist (list of string): list of path to the
            template directory in CAPS folder for each participant, for
            the participants that have already been processed
        out_overwrite_tsv (str): same as in_rundir. Path to the .tsv to
            use if the user wishes to overwrite the subjects that have
            already been processed in the CAPS folder. None if no
            subjects already processed.
    """
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils

    # sanity check: check at least one subject
    if not in_subject_list:
        error_msg = 'Error: there should be at least one subject'
        error_msg = '{0} to conduct the longitudinal study'.format(error_msg)
        raise ValueError(error_msg)

    # check if we have the necessary (cross-sectional) recon-all data
    # prior to running the template creation. Will crash if we do not
    # have this.
    utils.check_xsectional_reconalled(
        in_caps_dir, in_subject_list, in_session_list)

    # Check target CAPS target dir for template creation before
    # conducting any processing
    [
        pcssd_sublist,
        pcssd_seslist2,
        out_pcssd_capstargetlist,
        out_unpcssd_sublist,
        unpcssd_seslist2,
        unpcssd_capstargetlist,
        all_sublist,
        all_seslist2,
        all_capstargetlist] = utils.check_caps_template(
            in_subject_list, in_session_list, in_caps_dir)

    # Deduce subjects to be processed next (depending on whether
    # '--force-overwrite' flag was provided or not)
    [
        out_subject_list,
        out_session_list2,
        out_capstarget_list,
        out_overwrite_tsv] = utils.to_process(
            pcssd_sublist,
            pcssd_seslist2,
            out_unpcssd_sublist,
            unpcssd_seslist2,
            unpcssd_capstargetlist,
            all_sublist,
            all_seslist2,
            all_capstargetlist,
            in_caps_dir,
            in_overwrite_caps,
            in_working_directory,
            in_n_procs,
            in_rundir)

    # Check if any subject can be found for which only one time point
    # has been found, in which case warn users of potential issues
    utils.check_single_timepoint(out_subject_list, out_session_list2)

    return [
        out_subject_list,
        out_session_list2,
        out_capstarget_list,
        out_unpcssd_sublist,
        out_pcssd_capstargetlist,
        out_overwrite_tsv]


def store_reconallbase_results(in_subject_list):
    """Create a folder to store all recon-all -base outputs

    This creates a 'store 'folder that will store, for each subject, the
    recon-all -base output. The function returns the path to the store
    folder.
    The folder is initially empty. At a later stage, a MapNode is run
    on each subject to carry out the recon-all -base in a temp folder
    and when completed, send the result to this store folder. This
    way, the map node can be re-run (in case it crashed) and won't have
    to re-compute the template if already done for any particular
    subject.

    Args:
        in_subject_list (list): list of participant_id. This is only
            used to connect the input_node to the store_reconall_node
            and serves no other purpose within the current function.

    Returns:
        out_workdirstore_path (string): path to the 'store' folder that
            will contain the output for recon-all -base on all subjects
            when the command is run in a subsequent node
    """
    import os
    import errno

    # create subject-specific path in Clinica's working directory
    # for the FreeSurfer 'recon-all' commands
    out_workdirstore_path = os.path.abspath("./subjects")
    try:
        os.mkdir(out_workdirstore_path)
    except OSError as oserror:
        if oserror.errno != errno.EEXIST:
            raise

    return out_workdirstore_path


def get_reconallbase_flags(in_subject, in_session_list):
    """Create reconall -base flags

    Creates the flags that are required to run the recon-all -base for
    each subject (i.e. to create the unbiased template for all time
    points for the subject).

    Args:
        in_subject (string): (unique) participant_id
        in_session_list (list of string): list of session_id
            corresponding to a specific participant_id containing all
            session symlinks for any particular subject_id

    Returns:
        out_reconallbase_flags (string): all the flags to run
            recon-all -base for any particular subject and their
            associated list of sessions (time points)
    """
    # template ID: subject name
    out_reconallbase_flags = '-base {0}'.format(in_subject)

    # timepoints
    for session in in_session_list:
        sub_ses = '{0}_{1}'.format(in_subject, session)
        out_reconallbase_flags = '{0} -tp {1}'.format(
            out_reconallbase_flags, sub_ses)

    # -all directive
    out_reconallbase_flags = '{0} -all'.format(out_reconallbase_flags)

    return out_reconallbase_flags


def check_reconall_base_single(subjects_dir, in_subject):
    """Check if recon-all base run successfully for any single subject

    Args:
        subjects_dir (string): CAPS subdirectory containing the subject
            (absolute path)
        in_subject (string): participant_id

    Returns:
        out_template_created (Boolean): True if success, False otherwise
    """
    import os
    import subprocess

    # check if longitudinal base subfolder exists
    subjectid_longitudinal = "{0}/{1}".format(subjects_dir, in_subject)
    if not os.path.isdir(subjectid_longitudinal):
        error_msg = 'Error: {0} does not exist'.format(subjectid_longitudinal)
        raise IOError(error_msg)

    # check the recon-all.log
    log_file = os.path.join(subjectid_longitudinal, 'scripts', 'recon-all.log')
    if os.path.isfile(log_file):
        last_line = subprocess.check_output(['tail', '-1', log_file])
        if b'finished without error' not in last_line:
            out_template_created = False
            error_msg = 'Subject {0} has not been processed yet'.format(
                in_subject)
            error_msg = '{0}. Re-run t1-freesurfer-template.'.format(error_msg)
            raise IOError(error_msg)
        else:
            out_template_created = True
    else:
        error_msg = 'Error: {0} does not exist'.format(log_file)
        raise IOError(error_msg)

    return out_template_created


def create_fssubdir_path(subject, session_list):
    """Create a subdir to FS recon-all

    This creates a subdir that will contain the outpute of FreeSurfer
    'recon-all -base' command. Distinguishes between two cases:
    1) only one session is linked to the subject
    - there currently (as of 22 Feb 2019) is a bug in Freesurfer
    recon-all -base, which in some cases (e.g., only one time point),
    will crash as it's trying to write lines too long for the shell to
    handle. This is caused by the path to FreeSurfer SUBJECT_DIR being
    too long itself. The current function works around this issue by
    checking if there only is one session associated to a subject, and
    in that case, putting the SUBJECT_DIR inside the system temporary
    folder (Linux: /tmp, OS X: $TMPDIR) so that its path is as short as
    possible.
    - This causes an issue as sytem temporary folders are automatically
    cleaned up (e.g., every time the computer is rebooted or at
    periodical instances). In that case, a crash within the node
    corresponding to the current function means we can only re-run the
    pipeline safely if we re-do the recon-all -base computation in a new
    temporary location
    2) More than one session are linked to the subject: creates the
    temporary folder inside the working directory.
    The subdir is eventually moved to another store location (in
    function run_reconallbase), so that folders for both case 1) and
    case 2) are ultimately stored in the same place

    Args:
        subject (string): subject ID
        session_list (list of string): list of sessions ID associated to
            the current subject

    Returns:
        fssubdir_path (string): path to the subdir containing the output
            of 'recon-all -base'.
    """
    import os
    import errno
    import tempfile

    if len(session_list) == 1:
        # create FS subdir in temporary folder
        fssubdir_path = tempfile.mkdtemp()
    else:
        # create FS subdir in working dir, current node
        fssubdir_path = os.path.join(os.path.abspath('./'), subject)
        # create folder if not exists (else recon-all computations will
        # be resumed in the existing folder)
        try:
            os.mkdir(fssubdir_path)
        except OSError as oserror:
            if oserror.errno != errno.EEXIST:
                raise

    return fssubdir_path


def run_reconallbase(
        in_caps_dir,
        in_subject,
        in_session_list,
        in_reconallbase_flags,
        in_workdirstore_path):
    """Run recon-all -base function

    Run freesurfer template creation for each unique subject and their
    associated list of sessions. The recon-all command is built with
    previously computed arguments (flags).

    Args:
        in_caps_dir (string): CAPS directory to contain the output
        in_subject (string): participant_id
        in_session_list (list of string): list of all the sessions
            corresponding to the subject
        in_reconallbase_flags (string): all the flags to run
            recon-all -base for any particular subject and their
            associated list of sessions (time points)
        in_workdirstore_path (string): path to the 'store' folder that
            contain the output for recon-all -base on all subjects
            after the command has been run

    Returns:
        out_template_created (Boolean): True if successful, False
            otherwise
    """
    import os
    import errno
    import subprocess
    import shutil
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils

    # check if current subject has been processed already
    # (i.e., if corresponding folder already stored to working dir)
    workdir_subject_path = "{0}/{1}".format(in_workdirstore_path, in_subject)
    if not os.path.isdir(workdir_subject_path):
        # Hasn't been processed, or crashed before being transferred to
        # the working directory. We re-run recon-all -base

        # create temporary folder (store path in .txt file for
        # traceability) and insert a FreeSurfer SUBJECT_DIR there
        fssubdir_path = utils.create_fssubdir_path(in_subject, in_session_list)
        storepath_filename = os.path.abspath(
            "{0}/{1}_store_path.txt".format(in_workdirstore_path, in_subject))
        # Note: we might want to check first if file already exists and
        #       send a warning accordingly
        with open(storepath_filename, 'w') as storepath_file:
            storepath_file.write(fssubdir_path)

        # add symlinks to all subject_section cross-sectional run in the
        # FreeSurfer SUBJECT_DIR (temporary folder)
        caps_path = os.path.expanduser(in_caps_dir)
        caps_dir = os.path.join(caps_path, 'subjects')
        for session in in_session_list:
            # retrieve path to {subject,session} cross-sectional reconall
            # runs
            caps_subses_path = "{0}/{1}/{2}/t1/freesurfer_cross_sectional/{1}_{2}".format(
                caps_dir, in_subject, session)
            # define symlink path
            symlink_subses_path = "{0}/{1}_{2}".format(
                fssubdir_path, in_subject, session)
            if not os.path.exists(symlink_subses_path):
                os.symlink(caps_subses_path, symlink_subses_path)

        # run recon-all -base in temp FreeSurfer SUBJECT_DIR
        reconallbase_command = 'recon-all {0} -sd {1}'.format(
            in_reconallbase_flags, fssubdir_path)
        subprocess_run_reconallbase = subprocess.run(
            reconallbase_command,
            shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        if subprocess_run_reconallbase.returncode != 0:
            raise ValueError('recon-all -base failed, returned non-zero code')

        # check if completed successfully
        out_template_created = utils.check_reconall_base_single(
            fssubdir_path, in_subject)
        if out_template_created:
            # create subject-specific workdir store location for
            # recon-all -base results
            workdirstore_subject_path = os.path.join(
                in_workdirstore_path, in_subject)
            try:
                os.mkdir(workdirstore_subject_path)
            except OSError as oserror:
                if oserror.errno != errno.EEXIST:
                    raise
            # move the recon-all -base output to working directory
            list_filefolder = os.listdir(fssubdir_path)
            for entry in list_filefolder:
                shutil.move(
                    '{0}/{1}'.format(fssubdir_path, entry),
                    workdirstore_subject_path)
        else:
            # return an error
            error_msg = 'Error: recon-all -base did not complete.'
            error_msg = '{0} Template not created.'.format(error_msg)
            raise IOError(error_msg)

        # remove temporary folder
        # for now do nothing, the associated system temporary folder
        # should be empty any way...

    return out_template_created


def safer_rmtree(caps_target, caps_dir):
    """Only remove CAPS target if it's as subfolder of CAPS dir

    Will check if the CAPS target folder is a subfolder of CAPS dir
    prior to deleting it.
    This is to reduce the possibility of deleting files outside the CAPS
    dir in case of an accidental erroneous definition of the CAPS target
    in code.

    Args:
        caps_target (string): path to CAPS target subfolder for the
            template connected to the current subject
        caps_dir (string): CAPS directory to contain the output
            Used to make sure the folder we are overwriting (if
            overwrite-caps option provided) are subolders of the input
            CAPS dir

    Returns:
        N/A
    """
    import os
    import shutil

    # get absolute path + real (i.e., actual path for symlinks)
    caps_target_real = os.path.realpath(caps_target)
    caps_dir_real = os.path.realpath(caps_dir)
    caps_dir_real_sep = "{0}{1}".format(caps_dir_real, os.sep)

    # check if the path to image encompasses the path to folder
    if not caps_target_real.startswith(caps_dir_real_sep):
        error_msg = 'Error: CAPS target {0}'.format(caps_target)
        error_msg = '{0} is not a subfolder of caps_dir {1}.'.format(
            error_msg, caps_dir)
        error_msg = '{0} Cannot safely remove.'.format(error_msg)
        raise IOError(error_msg)

    # remove the caps target
    if os.path.exists(caps_target_real):
        shutil.rmtree(caps_target_real)
    else:
        error_msg = 'Error: folder {0} does not exist'.format(caps_target)
        raise IOError(error_msg)


def copy_to_caps(
        in_subject,
        in_subject_dir,
        in_caps_target,
        in_caps_dir,
        in_overwrite_caps,
        in_template_created):
    """Copy template folder to CAPS dir

    Move the template folder created by recon-all base to its correct
    location in the CAPS tree structure.

    Args:
        in_subject (string): (unique) participant_id
        in_subject_dir (string): location of the template folder (in the
            working directory)
        in_caps_target (string): path to CAPS target subfolder for the
            template connected to the current subject
        in_caps_dir (string): CAPS directory to contain the output
            Used to make sure the folder we are overwriting (if
            overwrite-caps option provided) are subolders of the input
            CAPS dir
        in_overwrite_caps (string): Option provided by the user to state
            whether the data in the CAPS folder (i.e.,
            previously-computed template) should be overwritten or not
            by a new run. By default, do not overwrite. Will overwrite
            for the following values: 'true', 'True' and 'TRUE'
        in_template_created (Boolean): Used to force template
            recon-all -base to be run before moving anything to the CAPS
            directory

    Returns:
        out_copy2_caps (boolean): flag to indicate whether or not the
            copy was successful
    """
    import os
    import shutil
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils

    # Convert overwrite-caps flag to boolean
    force_overwrite = utils.boolean_overwrite_caps(in_overwrite_caps)

    # define template location in the working directory
    # (repeating twice in_subject because in_subject_dir/in_subject
    # contains both in_subject template and cross-sectional
    # symlinks)
    wd_template_location = os.path.join(in_subject_dir, in_subject, in_subject)

    # Check if overwrite-caps option provided. If provided, delete the
    # target subdirectory in CAPS folder
    if force_overwrite:
        # check the target subdirectory exists in CAPS
        if os.path.exists(in_caps_target):
            # remove
            utils.safer_rmtree(in_caps_target, in_caps_dir)

    # copy the template from working directory to CAPS directory
    shutil.copytree(wd_template_location, in_caps_target)
    out_copy2_caps = True

    return out_copy2_caps


def sendto_longcorr(
        in_copy2_caps,
        in_unpcssd_sublist,
        in_pcssd_capstargetlist,
        in_overwrite_tsv):
    """Identity function

    Used to store the data from the template pipeline to be sent to the
    longitudinal pipeline.

    Args:
        in_copy2_caps (Boolean): Force the copy2caps_node to be run
            before sending data to longitudinal-correction (i.e., before
            starting the longitudinal-correction pipeline)
        in_unpcssd_sublist (list of string): list of participant_id,
            where each participant appears only once, wich have not
            already been (template) processed
        in_pcssd_capstargetlist (list of string): list of path to the
            template directory in CAPS folder for each participant, for
            the participants that have already been processed
        in_overwrite_tsv (str): path to the .tsv file to use if the user
            wishes to re-run the pipeline on subjects that had already
            been processed. None if no subjects already processed.

    Returns:
        out_unpcssd_sublist (list of string): same as in_unpcssd_sublist
        out_pcssd_capstargetlist (list of string): same as
            in_pcssd_capstargetlist
        out_overwrite_tsv (str): same as in_overwrite_tsv
    """
    out_unpcssd_sublist = in_unpcssd_sublist
    out_pcssd_capstargetlist = in_pcssd_capstargetlist
    out_overwrite_tsv = in_overwrite_tsv

    return out_unpcssd_sublist, out_pcssd_capstargetlist, out_overwrite_tsv



# coding: utf8

import clinica.engine as ce


class fMRIPreprocessingCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 'fmri-preprocessing'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Preprocessing of raw fMRI datasets:\n'
                             'http://clinica.run/doc/Pipelines/fMRI_Preprocessing/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')

        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        default_fwhm = [8, 8, 8]
        optional.add_argument("-fwhm", "--full_width_at_half_maximum",
                              metavar="FWHM",
                              nargs=3, type=int, default=default_fwhm,
                              help="A list of integers specifying the full width at half maximum (FWHM) filter size "
                                   "in the three directions x y z, in millimeters, used to smooth the images "
                                   "(default: --full_width_at_half_maximum %s)." %
                                   self.list_to_string(default_fwhm))
        optional.add_argument("-t1s", "--t1_native_space",
                              action='store_true', default=False,
                              help="When specified, the corrected fMRI data registered to the native T1 space "
                                   "of each subject are saved. This might be useful when dealing with "
                                   "parcellations that are in the native T1 space.")
        optional.add_argument("-fsbm", "--freesurfer_brain_mask",
                              action='store_true', default=False,
                              help="When specified, the pipeline uses the brain extraction generated by FreeSurfer "
                                   "during the t1-freesurfer pipeline. Otherwise, it computes it using the "
                                   "SPM Registration tool followed by empirically-estimated operations "
                                   "performed by FSL.")
        optional.add_argument("-u", "--unwarping",
                              action='store_true', default=False,
                              help="When specified, the unwarping correction step is executed. This will add "
                                   "SPM's Unwarping to the Realign step (magnitude and phasediff files are necessary "
                                   "in the BIDS directory).")

        # Clinica standard arguments (e.g. --n_procs)
        clinica_opt = self.add_clinica_standard_arguments()
        # clinica_opt.add_argument("-sl", "--slurm", action='store_true',
        #                          help='Run the pipelines using SLURM.')
        # clinica_opt.add_argument("-sa", "--sbatch_args",
        #                          help='SLURM\'s sbatch tool arguments.')

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from networkx import Graph
        from .fmri_preprocessing_pipeline import fMRIPreprocessing
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit

        parameters = {
            'full_width_at_half_maximum': args.full_width_at_half_maximum,
            't1_native_space': args.t1_native_space,
            'freesurfer_brain_mask': args.freesurfer_brain_mask,
            'unwarping': args.unwarping,
        }
        pipeline = fMRIPreprocessing(
            bids_directory=self.absolute_path(args.bids_directory),
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        # elif args.slurm:
        #     exec_pipeline = pipeline.run(plugin='SLURMGraph',
        #                                  plugin_args={'dont_resubmit_completed_jobs': True,
        #                                               'sbatch_args': args.sbatch_args})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf-8


def perform_gtmseg(caps_dir, subject_id, session_id):
    """gtmseg is a freesurfer command used to perform a segmentation used in some partial volume correction methods.

    Warnings:
        - This method changes the environment variable $SUBJECTS_DIR (but put
          the original one back after execution).  This has not been intensely
          tested whether it can lead to some problems : (for instance if 2
          subjects are running in parallel)

    Args:
        (string) caps_dir : CAPS directory.
        (string) subject_id: The subject_id (something like sub-ADNI002S4213)
        (string) session_id: The session id ( something like : ses-M12)

    Returns:
        (string) Path to the segmentation volume : a volume where each voxel
        has a label (ranging [0 2035] ), see Freesurfer lookup table to see the
        labels with their corresponding names.
    """
    import os
    import shutil
    import nipype.pipeline.engine as pe
    from clinica.utils.stream import cprint
    from nipype.interfaces.base import CommandLine

    # Old subject_dir is saved for later
    subjects_dir_backup = os.path.expandvars('$SUBJECTS_DIR')
    # Set the new subject dir for the function to work properly
    os.environ["SUBJECTS_DIR"] = os.path.join(caps_dir,
                                              'subjects',
                                              subject_id,
                                              session_id,
                                              't1',
                                              'freesurfer_cross_sectional')
    if not os.path.exists(os.path.join(os.path.expandvars('$SUBJECTS_DIR'), subject_id + '_' + session_id, 'mri', 'gtmseg.mgz')):
        # Creation of standalone node based on Command Line Interface.
        # We simply put the command line we would run on a console
        segmentation = pe.Node(interface=CommandLine('gtmseg --s %s_%s --no-seg-stats --xcerseg' % (subject_id, session_id),
                                                     terminal_output='stream'), name='gtmseg')
        segmentation.run()

    # We specify the out file to be in the current directory of execution (easy for us to look at it afterward in the
    # working directory). We copy then the file.
    out_file = os.path.abspath('./gtmseg.mgz')
    shutil.copy(os.path.join(os.path.expandvars('$SUBJECTS_DIR'), subject_id + '_' + session_id, 'mri', 'gtmseg.mgz'),
                out_file)

    # Remove bunch of files created during segmentation in caps dir and not needed
    gtmsegcab = os.path.join(os.path.expandvars('$SUBJECTS_DIR'), subject_id + '_' + session_id, 'mri', 'gtmseg.ctab')
    if os.path.exists(gtmsegcab):
        os.remove(gtmsegcab)

    gtmseglta = os.path.join(os.path.expandvars('$SUBJECTS_DIR'), subject_id + '_' + session_id, 'mri', 'gtmseg.lta')
    if os.path.exists(gtmseglta):
        os.remove(gtmseglta)

    # Set back the SUBJECT_DIR environment variable of the user
    os.environ["SUBJECTS_DIR"] = subjects_dir_backup
    return out_file


def remove_nan(volname):
    """remove_nan is a method needed after a registration performed by spmregister : insted of filling space with 0, nan
    are used to extend the PET space. We propose to replace them with 0s.

    Args:
        (string) volname : path to the Nifti volume where NaNs need to be replaced by 0s

    Returns:
        (string) Path to the volume in Nifti that does not contain any NaNs
    """
    import nibabel as nib
    import os

    # Load the volume and get the data
    nifti_in = nib.load(volname)
    data = nifti_in.get_data()

    # Extract mask where input data is different from itself (equivalent to finding NaNs, because NaN != NaN is True)
    # and replace with 0s
    nan_mask = data != data
    data[nan_mask] = 0

    # Now create final image (using header of original image), and save it in current directory
    nifti_out = nib.Nifti1Image(data, nifti_in.affine, header=nifti_in.header)
    filename = os.path.basename(volname)
    vol_wo_nan = './no_nan_' + filename + '.gz'
    vol_wo_nan = os.path.abspath(vol_wo_nan)
    nib.save(nifti_out, vol_wo_nan)
    return vol_wo_nan


def make_label_conversion(gtmsegfile, csv):
    """make_label_conversion is a method used on the segmentation from gtmsegmentation. The purpose is to reduce the
    number of label. The gathering of labels is specified in a separate file

    Args:
        (string) gtmsegfile   : path to the Nifti volume containing the gtmseg segmentation
        (string) csv          : path to .csv file that contains 3 columns : REGION SOURCE DST. Separator is , (coma).

    Returns:
        (list of strings) List of path to the converted volumes according to the .csv file. Each volume is a mask
        representing an area
    """
    import nibabel as nib
    import numpy
    import os
    import pandas
    from clinica.utils.stream import cprint

    def isclose(a, b, rel_tol=1e-9, abs_tol=0.0):
        """ Small function designed to measure equality between to floating or double numbers, using 2 thresholds : a
        relative tolerance, and an absolute tolerance
        """
        return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)

    # Read label from gtmsegfile, change data into integers in order to have no problems when testing equality of labels
    label = nib.load(gtmsegfile)
    label.header.set_data_dtype('int8')
    volume = label.get_data()

    # Unique function gives a list where each label of the volume is listed once
    old_labels = numpy.unique(volume)
    old_labels = old_labels.astype('int16')

    # allsum is a control volume (sum of a pixel across the 4 th dimension must be equal to 1)
    allsum = numpy.zeros(volume.shape)

    # Reading of csv file, raise exception if the pattern REGION, SOURCE, DST is not found
    if not os.path.isfile(csv):
        raise Exception('The CSV file does not exist.')
    convert_lut = pandas.io.parsers.read_csv(csv, sep=',')
    if list(convert_lut.columns.values) != ['REGION', 'SOURCE', 'DST']:
        raise Exception('CSV file ' + csv + ' is not in the correct format. Columns should be : REGION, SOURCE, DST')

    # Extract columns to a list form (values converted into integers)
    src = list(convert_lut.SOURCE)
    src_val = numpy.asanyarray(src)
    src_val = src_val.astype('int')

    dst = list(convert_lut.DST)
    dst_val = numpy.asarray(dst)
    dst_val = dst_val.astype('int')

    reg = list(convert_lut.REGION)

    # Check that each label of original volume (old_label) has a matching transformation in the csv file
    for i in range(old_labels.size):
        index = numpy.argwhere(src_val == old_labels[i])
        # Size 0 means no occurence found
        if index.size == 0:
            raise Exception('Could not find label '
                            + str(old_labels[i])
                            + ' on conversion table. Add it manually in csv file to correct error')
        else:
            cprint(str(old_labels[i]) + ' has been found on conversion table')

    # Instanciation of final volume, with same dtype as original volume
    new_volume = numpy.zeros(volume.shape, dtype=volume.dtype)
    # Computing the transformation
    for i in range(len(src)):
        new_volume[volume == src_val[i]] = dst_val[i]
        cprint('Region ' + reg[i] + ' transformed')
    # Get unique list of new label
    new_labels = numpy.unique(new_volume)
    new_labels = new_labels.astype('int')
    list_of_regions = list()

    # For each label, create a volume file filled with 0s and 1s and save it in current directory under whatever name
    for i in range(new_labels.size):
        region_volume = numpy.zeros(volume.shape, dtype='uint8')
        region_volume[new_volume == new_labels[i]] = 1
        myNifti = nib.Nifti1Image(region_volume, label.affine, header=label.header)
        current_path = './' + str(new_labels[i]) + '.nii.gz'
        current_path = os.path.abspath(current_path)
        nib.save(myNifti, current_path)
        list_of_regions.append(current_path)
        cprint('Label ' + str(new_labels[i]) + ' created in ' + current_path)
        allsum = allsum + region_volume

    # The sum of a voxel location across the fourth dimension should be 1
    sum_voxel_mean = float(sum(sum(sum(allsum)))) / allsum.size
    if not isclose(1.0, sum_voxel_mean):
        raise Exception('Problem during parcellation : the mean sum of a voxel across 4th dimension is '
                        + str(sum_voxel_mean)
                        + ' instead of 1.0')
    # The list of files is returned
    return list_of_regions


def runApplyInverseDeformationField_SPM_standalone(target,
                                                   deformation_field,
                                                   img,
                                                   matscript_folder):
    """
    This function does the exact same job as runApplyInverseDeformationField but with SPM standalone. We directly create
    a batch file that SPM standalone can run. This function does not check wether SPM standalone must be used. Previous
    check when building the pipeline ensures that all the env vars exists ($SPMSTANDALONE_HOME and $MCR_HOME)
    """
    from os.path import abspath, join, basename, exists
    import platform
    import os

    prefix = 'subject_space_'

    # Write SPM batch command directly in a script that is readable by SPM standalone
    script_location = abspath('./m_script.m')
    script_file = open(script_location, 'w+')
    script_file.write('jobs{1}.spm.util.defs.comp{1}.inv.comp{1}.def = {\'' + deformation_field + '\'};\n')
    script_file.write('jobs{1}.spm.util.defs.comp{1}.inv.space = {\'' + target + '\'};\n')
    script_file.write('jobs{1}.spm.util.defs.out{1}.pull.fnames = {\'' + img + '\'};\n')
    script_file.write('jobs{1}.spm.util.defs.out{1}.pull.savedir.saveusr = {\'' + abspath(os.getcwd()) + '\'};\n')
    script_file.write('jobs{1}.spm.util.defs.out{1}.pull.interp = 4;\n')
    script_file.write('jobs{1}.spm.util.defs.out{1}.pull.mask = 1;\n')
    script_file.write('jobs{1}.spm.util.defs.out{1}.pull.fwhm = [0 0 0];\n')
    script_file.write('jobs{1}.spm.util.defs.out{1}.pull.prefix = \'' + prefix + '\';\n')
    script_file.close()

    # Generate command line to run
    # SPM standalone must be run directly from its root folder
    if platform.system() == 'Darwin':
        # Mac OS
        cmdline = 'cd $SPMSTANDALONE_HOME && ./run_spm12.sh $MCR_HOME batch ' + script_location
    elif platform.system() == 'Linux':
        # Linux OS
        cmdline = '$SPMSTANDALONE_HOME/run_spm12.sh $MCR_HOME batch ' + script_location
    else:
        raise SystemError('Only support Mac OS and Linux')
    os.system(cmdline)

    output_file = join(abspath('./'), prefix + basename(img))
    if not exists(output_file):
        raise IOError('Something went wrong while trying to run runApplyInverseDeformationField_SPM_standalone'
                      + '. Output file not generated. Command launched :\n\t '
                      + cmdline + '\n. We strongly recommand that you use the supported version of Matlab MCR '
                      + ' recommanded by the creators of SPM.')
    return output_file


def runApplyInverseDeformationField(target,
                                    deformation_field,
                                    img,
                                    matscript_folder):

    import sys
    import os
    from nipype.interfaces.matlab import MatlabCommand, get_matlab_command

    prefix = 'subject_space_'

    MatlabCommand.set_default_matlab_cmd(get_matlab_command())
    matlab = MatlabCommand()
    if sys.platform.startswith('linux'):
        matlab.inputs.args = '-nosoftwareopengl'

    matlab.inputs.paths = matscript_folder
    matlab.inputs.script = """
    applyInverseDeformationField('%s', '%s', '%s', './', '%s')
    """ % (target, deformation_field, img, prefix)
    matlab.inputs.single_comp_thread = False
    matlab.inputs.logfile = os.path.join('./', 'matlab_output.log')
    matlab.run()

    output_file = os.path.join(os.path.abspath('./'), prefix + os.path.basename(img))
    if not os.path.exists(output_file):
        raise IOError('Something went wrong, please check ' + os.path.abspath(matlab.inputs.logfile) + ' for more information')
    return output_file


def suvr_normalization(pet_path, mask):
    """suvr_normalization is a way of getting suvr from your pet image, based on the segmentation performed by
    gtmsegmentation. The Standard Uptake Value ratio is computed by dividing the whole PET volume by the mean value
    observed in the pons.

    Args:
        (string) pet_path     : path to the Nifti volume containing PET scan, realigned on upsampled T1
        (string) mask         : mask of the pons (fdg) or pons+cerebellum (av45) already eroded

    Returns:
        (string) Path to the suvr normalized volume in the current directory
    """
    import nibabel as nib
    import numpy
    import os

    # Load mask
    eroded_mask_nifti = nib.load(mask)
    eroded_mask = eroded_mask_nifti.get_data()
    eroded_mask = eroded_mask > 0

    # Load PET data (they must be in gtmsegspace, or same space as label file)
    pet = nib.load(pet_path)
    pet_data = pet.get_data()

    # check that eroded mask is not null
    mask_size = sum(sum(sum(eroded_mask)))
    if mask_size == 0:
        raise Exception('Number of non-zero value of mask is 0. A problem occured when moving the eroded mask from MNI to gtmsegspace')

    # Mask unwanted values to determine mean uptake value
    pons_pet_activity = eroded_mask * pet_data
    mean_pons_pet_activity = sum(sum(sum(pons_pet_activity))) / mask_size

    # Then normalize PET data by this mean activity
    suvr_pet_data = pet_data / mean_pons_pet_activity
    suvr = nib.Nifti1Image(suvr_pet_data, pet.affine, header=pet.header)
    suvr_filename = 'suvr_' + os.path.basename(pet_path)
    suvr_filename = os.path.abspath('./' + suvr_filename)
    nib.save(suvr, suvr_filename)
    return suvr_filename


def suvr_normalization2(pet_path, gtmseg_path, pet_tracer):
    """OLD VERSION OF SUVR_NORMALIZATION - NOT USED ANYMORE
    suvr_normalization is a way of getting suvr from your pet image, based on the segmentation performed by
    gtmsegmentation. The Standard Uptake Value ratio is computed by dividing the whole PET volume by the mean value
    observed in the pons. The mask of the pons is eroded 3 times by a kernel of 3x3x3, in order to have a more
    homogeneous area. In the case of AV45 PET, normalization is realized on pons and cerebellum.

    Args:
        (string) pet_path     : path to the Nifti volume containing PET scan, realigned on upsampled T1 whose
            segmentation is given in the last argument
        (string) gtmseg_path  : path to gtmsegmentation Nifti volume. At this point, the 2 inputs arguments must have
            the same voxel number and be realigned between each other
        (string) pet_tracer     : type of PET (fdg or av45). Normalization area varies across different markers.

    Returns:
        (string) Path to the suvr normalized volume in the current directory
    """
    import nibabel as nib
    import numpy
    from scipy import ndimage
    import os

    # Load labels
    label = nib.load(gtmseg_path)
    label_data = label.get_data()

    # Load PET data (they must be in gtmsegspace, or same space as label file)
    pet = nib.load(pet_path)
    pet_data = pet.get_data()

    if pet_tracer == 'fdg':
        # Extract Pons (label 174)
        normalization_mask = (label_data == 174)
    elif pet_tracer == 'av45':
        # Extract Pons + Cerebellum
        normalization_mask = (label_data == 174) | (label_data == 7) | (label_data == 8) | (label_data == 46) | (label_data == 47)
    else:
        raise Exception(pet_tracer + ' PET tracer not recognized. Must be either fdg or av45')

    # Use [3x3x3] structuring element filled with 1 to erode
    el = ndimage.generate_binary_structure(3, 3)
    el.astype(numpy.int)
    N_iter = 5
    eroded_normalization_mask = ndimage.binary_erosion(normalization_mask, el, N_iter)

    # check that eroded mask is not null
    mask_size = sum(sum(sum(eroded_normalization_mask)))
    while mask_size < 1:
        N_iter -= 1
        eroded_normalization_mask = ndimage.binary_erosion(normalization_mask, el, N_iter)
        mask_size = sum(sum(sum(eroded_normalization_mask)))
        if N_iter == 0 and mask_size == 0:
            raise Exception('Could not find reference structure in ' + gtmseg_path)

    # Mask non-Pons values to determine mean uptake value
    pons_pet_activity = eroded_normalization_mask * pet_data
    mean_pons_pet_activity = sum(sum(sum(pons_pet_activity))) / sum(sum(sum(eroded_normalization_mask)))

    # Then normalize PET data by this mean activity
    suvr_pet_data = pet_data / mean_pons_pet_activity
    suvr = nib.Nifti1Image(suvr_pet_data, pet.affine, header=pet.header)
    suvr_filename = 'suvr_' + os.path.basename(pet_path)
    suvr_filename = os.path.abspath('./' + suvr_filename)
    nib.save(suvr, suvr_filename)
    return suvr_filename


def mris_expand(in_surface):
    """mris_expand is using the freesurfer function of the same name. It expands the white input surface toward the pial,
    generating 7 surfaces at 35%, 40%, 45%, 50%, 55%, 60%, 65% of thickness.

    Args:
        (string) in_surface : Path to the input white surface, but must be named lh.white or rh.white, and the folder
            containing the surface file must also have ?h.pial, ?.sphere, ?h.thickness (freesurfer surf folder)

    Returns:
        (list of strings) List of path to the generated surfaces
    """
    import os
    import sys
    import shutil

    # bug in mris_expand : you are not allowed to write the surfaces elsewhere than in the surf folder

    # -N is a hidden parameter (not documented) that allows the user to specify the number of surface generated between
    # source and final target surface. Here target is 65% of thickness, with 13 surfaces. Then we only keep the surfaces
    # we are interested in.

    out_file = in_surface + '_exp-'
    cmd = 'mris_expand -thickness -N 13 ' + in_surface + ' 0.65 ' + out_file
    # If system is MacOS, this export command must be run just before the mri_vol2surf command to bypass MacOs security
    if sys.platform == 'darwin':
        cmd = 'export DYLD_LIBRARY_PATH=$FREESURFER_HOME/lib/gcc/lib && ' + cmd

    os.system(cmd)

    # Move surface file to the current directory
    out_filelist = [out_file + '007',
                    out_file + '008',
                    out_file + '009',
                    out_file + '010',
                    out_file + '011',
                    out_file + '012',
                    out_file + '013']
    out_in_node = [os.path.abspath('./' + os.path.basename(x)) for x in out_filelist]
    for i in range(len(out_filelist)):
        shutil.move(out_filelist[i], out_in_node[i])

    # Remove useless surfaces (0%, 5%, 10%, 15%, 20%, 25% and 30% of thickness)
    to_remove = [out_file + '000',
                 out_file + '001',
                 out_file + '002',
                 out_file + '003',
                 out_file + '004',
                 out_file + '005',
                 out_file + '006']
    for i in range(len(to_remove)):
        os.remove(to_remove[i])

    return out_in_node


def surf2surf(in_surface, reg_file, gtmsegfile, subject_id, session_id, caps_dir):
    """surf2surf is a wrapper of freesurfer command mri_surf2surf. Here the aim is to convert a input surface (which is
    the native space of the subject), into the same surface but in the gtmseg space (space of the volume generated by
    the gtmsegmentation)

    Args:
        (string) in_surface : surface file that needs to be converted
        (string) reg_file   : Path to a registration file that represents the transformation needed to go from the native
            space to the gtmsegspace (see https://surfer.nmr.mgh.harvard.edu/fswiki/FsAnat-to-NativeAnat for more
            details)
        (string) gtmsegfile : Path to the gtm segmentation file
        (string) subject_id : The subject_id (something like sub-ADNI002S4213)
        (string) session_id : The session id ( something like : ses-M12)
        (string) caps_dir   : Path to the CAPS directory

    Returns:
        (string) Path to the converted surface in current directory
    """
    import os
    import shutil
    import sys

    # set subjects_dir env. variable for mri_surf2surf to work properly
    subjects_dir_backup = os.path.expandvars('$SUBJECTS_DIR')
    os.environ["SUBJECTS_DIR"] = os.path.join(caps_dir,
                                              'subjects',
                                              subject_id,
                                              session_id,
                                              't1',
                                              'freesurfer_cross_sectional')

    # make a copy of surface file to surface directory in CAPS in order to allow processing
    shutil.copy(in_surface, os.path.join(os.path.expandvars('$SUBJECTS_DIR'),
                                         subject_id + '_' + session_id,
                                         'surf'))

    # TODO write nicer way to grab hemi & filename (difficulty caused by the dots in filenames)
    # extract hemisphere based on filename
    hemi = os.path.basename(in_surface)[0:2]
    surfname = os.path.basename(in_surface)[3:]

    # Perform surf2surf algorithm
    tval = os.path.abspath('./' + os.path.basename(in_surface) + '_gtmsegspace')
    cmd = 'mri_surf2surf --reg %s %s --sval-xyz %s --hemi %s --tval-xyz %s --tval %s --s %s ' \
          % (reg_file, gtmsegfile, surfname, hemi, gtmsegfile, tval, subject_id + '_' + session_id)

    # If system is MacOS, this export command must be run just before the mri_vol2surf command to bypass MacOs security
    if sys.platform == 'darwin':
        cmd = 'export DYLD_LIBRARY_PATH=$FREESURFER_HOME/lib/gcc/lib && ' + cmd
    os.system(cmd)

    # remove file in caps
    os.remove(os.path.join(os.path.expandvars('$SUBJECTS_DIR'),
                           subject_id + '_' + session_id,
                           'surf',
                           os.path.basename(in_surface)))

    # put back original subjects_dir env
    os.environ["SUBJECTS_DIR"] = subjects_dir_backup

    return tval


def vol2surf(volume, surface, subject_id, session_id, caps_dir, gtmsegfile):
    """vol2surf is a wrapper of freesurfer command mri_vol2surf. It projects the volume into the surface : the value at
    each vertex is given by the value of the voxel it intersects

    Args:
        (string) volume     : Path to PET volume (in gtmseg space) that needs to be mapped into surface
        (string) surface    : Path to surface file
        (string) gtmsegfile :l Path to the gtm segmentation file (provides information on space, labels are not used
        (string) subject_id : The subject_id (something like sub-ADNI002S4213)
        (string) session_id : The session id ( something like : ses-M12)
        (string) caps_dir   : Path to the CAPS directory

    Returns:
        (string) Path to the data projected onto the surface
    """
    import os
    import shutil
    import sys

    # set subjects_dir env. variable for mri_vol2surf to work properly
    subjects_dir_backup = os.path.expandvars('$SUBJECTS_DIR')
    os.environ["SUBJECTS_DIR"] = os.path.join(caps_dir,
                                              'subjects',
                                              subject_id,
                                              session_id,
                                              't1',
                                              'freesurfer_cross_sectional')

    # TODO write nicer way to grab hemi & filename (difficulty caused by the dots in filenames)
    # extract hemisphere based on filename
    hemi = os.path.basename(surface)[0:2]
    surfname = os.path.basename(surface)[3:]

    # copy surface file in caps surf folder to allow processing
    shutil.copy(surface, os.path.join(os.path.expandvars('$SUBJECTS_DIR'),
                                      subject_id + '_' + session_id,
                                      'surf'))

    if not os.path.exists(os.path.join(os.path.expandvars('$SUBJECTS_DIR'),
                                       subject_id + '_' + session_id,
                                       'mri',
                                       'gtmseg.mgz')):
        shutil.copy(gtmsegfile, os.path.join(os.path.expandvars('$SUBJECTS_DIR'),
                                             subject_id + '_' + session_id,
                                             'mri',
                                             'gtmseg.mgz'))

    # execute vol2surf
    output = os.path.abspath('./' + hemi + '.projection_' + os.path.basename(surface) + '.mgh')
    cmd = 'mri_vol2surf'
    cmd += ' --mov ' + volume
    cmd += ' --o ' + output
    cmd += ' --surf ' + surfname
    cmd += ' --hemi ' + hemi
    cmd += ' --regheader ' + subject_id + '_' + session_id
    cmd += ' --ref gtmseg.mgz'
    cmd += ' --interp nearest'

    # If system is MacOS, this export command must be run just before the mri_vol2surf command to bypass MacOs security
    if sys.platform == 'darwin':
        cmd = 'export DYLD_LIBRARY_PATH=$FREESURFER_HOME/lib/gcc/lib && ' + cmd
    os.system(cmd)

    # remove file in caps
    os.remove(os.path.join(os.path.expandvars('$SUBJECTS_DIR'),
                           subject_id + '_' + session_id,
                           'surf',
                           os.path.basename(surface)))
    # TODO carreful here...
    # Removing gtmseg.mgz may lead to problems as other vol2surf are using it
    os.remove(os.path.join(os.path.expandvars('$SUBJECTS_DIR'),
                           subject_id + '_' + session_id,
                           'mri',
                           'gtmseg.mgz'))

    # put back original subjects_dir env
    os.environ["SUBJECTS_DIR"] = subjects_dir_backup

    return output


def weighted_mean(in_surfaces):
    """weighted_mean make a weighted average at each node of the surface. The weight are defined by a normal
    distribution (centered on the mid surface)

    Args:
        (list of strings) in_surfaces : List of path to the data projected on the 7 surfaces (35 to 65 % of thickness)
            at each nodes)

    Returns:
        (string) Path to the data averaged
    """
    import nibabel as nib
    import numpy as np
    import os

    # coefficient for normal repartition
    coefficient = [0.1034, 0.1399, 0.1677, 0.1782, 0.1677, 0.1399, 0.1034]

    # sample only to get dimension
    sample = nib.load(in_surfaces[0])
    data_normalized = np.zeros(sample.header.get_data_shape())

    if len(in_surfaces) != 7:
        raise Exception('There should be 7 surfaces at this point of the pipeline, but found '
                        + str(len(in_surfaces))
                        + ', something went wrong...')

    for i in range(len(in_surfaces)):
        current_surf = nib.load(in_surfaces[i])
        data_normalized += current_surf.get_data() * coefficient[i]

    # hemisphere name will always be in our case the first 2 letters of the filename
    hemi = os.path.basename(in_surfaces[0])[0:2]
    # data_normalized = np.atleast_3d(data_normalized)
    hemi_projection = nib.MGHImage(data_normalized, affine=sample.affine, header=sample.header)
    out_surface = './' + hemi + '.averaged_projection_on_cortical_surface.mgh'
    out_surface = os.path.abspath(out_surface)
    nib.save(hemi_projection, out_surface)

    return out_surface


def fsaverage_projection(projection, subject_id, caps_dir, session_id, fwhm):
    """fsaverage_projection projects your data into an averaged subject called fsaverage, available in your $SUBJECTS_DIR
    folder. fsaverage and the subject must be in the subject_dir, so a copy of fsaverage is performed if necessary

    Args:
        (string) projection : Path to the projected data onto native subject surface
        (string) subject_id : The subject id (something like sub-ADNI002S4213)
        (string) session_id : The session id ( something like : ses-M12)
        (string) caps_dir   : Path to the CAPS directory
        (float) fwhm        : FWHM of the Gaussian filter used for smoothing on fsaverage surface (not volume !)

    Returns:
        (string) Path to the data averaged
    """
    from nipype.interfaces.freesurfer import MRISPreproc
    import os
    import shutil

    subjects_dir_backup = os.path.expandvars('$SUBJECTS_DIR')
    os.environ["SUBJECTS_DIR"] = os.path.join(caps_dir,
                                              'subjects',
                                              subject_id,
                                              session_id,
                                              't1',
                                              'freesurfer_cross_sectional')

    # copy fsaverage folder next to : subject_id + '_' + session_id
    # for the mris_preproc command to properly find src and target
    fsaverage_has_been_copied = False
    if not os.path.exists(os.path.join(os.path.expandvars('$SUBJECTS_DIR'), 'fsaverage')):
        shutil.copytree(os.path.join(subjects_dir_backup, 'fsaverage'),
                        os.path.join(os.path.expandvars('$SUBJECTS_DIR'), 'fsaverage'))
        fsaverage_has_been_copied = True

    # also copy the mgh file in the surf folder (needed by MRISPreproc
    projection_in_surf_folder = os.path.join(os.path.expandvars('$SUBJECTS_DIR'),
                                             subject_id + '_' + session_id,
                                             'surf',
                                             os.path.basename(projection))

    if not os.path.exists(projection_in_surf_folder):
        shutil.copy(projection, projection_in_surf_folder)

    hemi = os.path.basename(projection)[0:2]
    out_fsaverage = os.path.abspath('./fsaverage_fwhm-' + str(fwhm) + '_' + os.path.basename(projection))

    # Use standalone node
    fsproj = MRISPreproc()
    fsproj.inputs.target = 'fsaverage'
    fsproj.inputs.subjects = [subject_id + '_' + session_id]
    fsproj.inputs.fwhm = fwhm
    fsproj.inputs.hemi = hemi
    fsproj.inputs.surf_measure = os.path.basename(projection)[3:]
    fsproj.inputs.out_file = out_fsaverage
    fsproj.run()

    # remove projection file from surf folder
    os.remove(projection_in_surf_folder)

    # remove fsaverage if it has been copied
    if fsaverage_has_been_copied:
        shutil.rmtree(os.path.join(os.path.expandvars('$SUBJECTS_DIR'), 'fsaverage'))

    # put back original subjects_dir env
    os.environ["SUBJECTS_DIR"] = subjects_dir_backup
    return out_fsaverage


def get_mid_surface(in_surfaces):
    """get_mid_surface gives the mid surface when dealing with the 7 different surfaces

        Args:
            (list of strings) in_surfaces : List of path to the 7 different surfaces generated by mris_expand

        Returns:
            (string) Path to the mid surface
        """
    return in_surfaces[3]


def readpsf(json_path):
    """readpsf is able to read the json file associated with the PET volume, and extract information needed to perform
    partial volume correction : the effective resolution in plane, and the effective resolution axial.

        Args:
            (string) json_path : Path to the json file containing the information

        Returns:
            (float) The effective resolution in plane
            (float) The effective axial resolution
        """
    import json
    import os

    if not os.path.exists(json_path):
        raise Exception('File ' + json_path + ' does not exist. Point spread function information must be provided.')

    # This is a standard way of reading data in a json
    with open(json_path) as df:
        data = json.load(df)
    in_plane = data['Psf'][0]['EffectiveResolutionInPlane']
    axial = data['Psf'][0]['EffectiveResolutionAxial']
    return in_plane, axial


def reformat_surfname(hemi, left_surface, right_surface):
    res = None
    if hemi == 'lh':
        res = left_surface
    elif hemi == 'rh':
        res = right_surface
    else:
        raise Exception('First input of this reformat_surfname function must be either lh or rh')
    return res


def normalization_areas(pet_tracer):
    if pet_tracer.upper() == 'FDG':
        return 'pons'
    elif pet_tracer.upper() == 'AV45':
        return 'pons-cerebellum'


def produce_tsv(pet, atlas_files):
    """produce_tsv computes the average of PET signal based on annot files from Freesurfer. Those files describes the
    brain according to known atlases.

        Args:
            (string) pet      : list of path to the PET projection (must be a MGH file) [left_hemisphere, right_hemisphere]
            (string) atlas_files  : Dictionnary containing path to lh and rh annotation files for any number of atlases.

        Returns:
            (string) tsv  : path to the tsv containing average PET values
        """
    import nibabel as nib
    import numpy as np
    import pandas as pds
    import os
    from clinica.utils.stream import cprint

    # Extract data from projected PET data
    lh_pet_mgh = np.squeeze(nib.load(pet[0]).get_data())
    rh_pet_mgh = np.squeeze(nib.load(pet[1]).get_data())

    filename_tsv = []
    for atlas in atlas_files:

        annot_atlas_left = nib.freesurfer.io.read_annot(atlas_files[atlas]['lh'], orig_ids=False)
        annot_atlas_left[0][annot_atlas_left[0] == -1] = 0
        annot_atlas_right = nib.freesurfer.io.read_annot(atlas_files[atlas]['rh'], orig_ids=False)
        annot_atlas_right[0][annot_atlas_right[0] == -1] = 0

        average_region = []
        region_names = []
        for r in range(len(annot_atlas_left[2])):
            cprint(annot_atlas_left[2][r])
            region_names.append(annot_atlas_left[2][r].astype(str) + '_lh')
            region_names.append(annot_atlas_left[2][r].astype(str) + '_rh')

            mask_left = annot_atlas_left[0] == r
            mask_left = np.uint(mask_left)

            masked_data_left = mask_left * lh_pet_mgh
            if np.sum(mask_left) == 0:
                average_region.append(np.nan)
            else:
                average_region.append(np.sum(masked_data_left) / np.sum(mask_left))

            mask_right = annot_atlas_right[0] == r
            mask_right = np.uint(mask_right)
            masked_data_right = mask_right * rh_pet_mgh
            if np.sum(mask_right) == 0:
                average_region.append(np.nan)
            else:
                average_region.append(np.sum(masked_data_right) / np.sum(mask_right))

        final_tsv = pds.DataFrame({'index': range(len(region_names)),
                                   'label_name': region_names,
                                   'mean_scalar': list(average_region)})
        filename_atlas_tsv = './' + atlas + '.tsv'
        filename_tsv.append(filename_atlas_tsv)
        final_tsv.to_csv(filename_atlas_tsv, sep='\t', index=False, columns=['index',
                                                                             'label_name',
                                                                             'mean_scalar'])
    return os.path.abspath(filename_tsv[0]), os.path.abspath(filename_tsv[1])


def get_wf(subject_id,
           session_id,
           psf,
           caps_dir,
           pet,
           orig_nu,
           white_surface_left,
           white_surface_right,
           working_directory_subjects,
           pet_tracer,
           csv_segmentation,
           subcortical_eroded_mask,
           matscript_folder_inverse_deformation,
           destrieux_left,
           destrieux_right,
           desikan_left,
           desikan_right,
           use_spm_standalone):
    """get_wf create a full workflow for only one subject, and then executes it

            Args:
                (string) subject_id          : The subject id
                (string) session_id          : The session id
                (string) psf                 : Path the json file containing information on the point spread function
                (string) caps_dir            : Path to the CAPS directory
                (string) pet                 : Path to the PET image in the bids directory
                (string) orig_nu             : Path to the orig_nu file (must be in the CAPS directory, in mri)
                (string) white_surface_left  : Path to the left white surface in native space of subject
                (string) white_surface_right : Path to the right white suface in native space of subject
                (string) csv_segmentation    : Path to the csv for the segmentation (problems encountered while using __file__)
                (string) subcortical_eroded_mask               : Path to the mask of the eroded version used for the suvr
                (string) matscript_folder_inverse_deformation  : Path to the current foler (containing the matlab script
                                                                 used to call the spm function for the inverse deformation

            Returns:
                Void
    """
    import os
    from clinica.utils.stream import cprint
    import nipype.pipeline.engine as pe
    import nipype.interfaces.utility as niu
    import nipype.interfaces.io as nio
    from nipype.algorithms.misc import Gunzip
    from nipype.interfaces.freesurfer import Tkregister2, ApplyVolTransform, MRIConvert
    from nipype.interfaces.fsl import Merge
    from nipype.interfaces.petpvc import PETPVC
    from nipype.interfaces.spm import Coregister, Normalize12
    import clinica.pipelines.pet_surface.pet_surface_utils as utils
    from clinica.utils.spm import get_tpm

    cprint('***** Beginning processing of ' + subject_id + ' on ' + session_id + ' *****')

    # Creation of workflow
    # 1 Creation of node

    unzip_pet = pe.Node(Gunzip(), name='unzip_pet')

    unzip_orig_nu = unzip_pet.clone(name='unzip_orig_nu')

    unzip_mask = unzip_pet.clone(name='unzip_mask')
    unzip_mask.inputs.in_file = subcortical_eroded_mask

    coreg = pe.Node(Coregister(), name='coreg')

    convert_mgh = pe.Node(MRIConvert(), name='convert_mgh')

    removenan = pe.Node(niu.Function(input_names=['volname'],
                                     output_names=['vol_wo_nan'],
                                     function=utils.remove_nan),
                        name='removenan')

    gtmsegmentation = pe.Node(niu.Function(input_names=['caps_dir',
                                                        'subject_id',
                                                        'session_id'],
                                           output_names=['gtmseg_file'],
                                           function=utils.perform_gtmseg),
                              name='gtmseg')
    gtmsegmentation.inputs.caps_dir = caps_dir
    gtmsegmentation.inputs.subject_id = subject_id
    gtmsegmentation.inputs.session_id = session_id

    tkregister = pe.Node(Tkregister2(reg_header=True), name='tkreg')

    convert_gtmseg = convert_mgh.clone(name='convert_gtmseg')

    labelconversion = pe.Node(niu.Function(input_names=['gtmsegfile',
                                                        'csv'],
                                           output_names=['list_of_regions'],
                                           function=utils.make_label_conversion),
                              name='conversion_of_labels')

    labelconversion.inputs.csv = csv_segmentation

    if not os.path.exists(labelconversion.inputs.csv):
        raise Exception('CSV file : ' + labelconversion.inputs.csv + ' does not exist.')

    merge_volume = pe.Node(Merge(output_type='NIFTI_GZ', dimension='t'),
                           name='merge_volume')

    vol2vol = pe.Node(ApplyVolTransform(reg_header=True, interp='trilin'),
                      name='vol2vol')

    vol2vol_mask = pe.Node(ApplyVolTransform(reg_header=True, interp='nearest'),
                           name='vol2vol_mask')

    tpm_nii = get_tpm()

    normalize12 = pe.Node(Normalize12(tpm=tpm_nii,
                                      affine_regularization_type='mni',
                                      jobtype='est',
                                      bias_fwhm=60,
                                      bias_regularization=0.0001,
                                      warping_regularization=[0, 0.001, 0.5, 0.05, 0.2]),
                          name='normalize_to_MNI')

    if use_spm_standalone is True:
        fun_apply_inverse_deformation = utils.runApplyInverseDeformationField_SPM_standalone
    else:
        fun_apply_inverse_deformation = utils.runApplyInverseDeformationField
    apply_inverse_deformation = pe.Node(niu.Function(input_names=['target',
                                                                  'deformation_field',
                                                                  'img',
                                                                  'matscript_folder'],
                                                     output_names=['freesurfer_space_eroded_mask'],
                                                     function=fun_apply_inverse_deformation),
                                        name='applyInverseDeformation')
    apply_inverse_deformation.inputs.matscript_folder = matscript_folder_inverse_deformation

    pons_normalization = pe.Node(niu.Function(input_names=['pet_path',
                                                           'mask'],
                                              output_names=['suvr'],
                                              function=utils.suvr_normalization),
                                 name='pons_normalization')
    pons_normalization.inputs.pet_tracer = pet_tracer

    pvc = pe.Node(PETPVC(pvc='IY'), name='petpvc')

    reformat_surface_name = pe.Node(niu.Function(input_names=['hemi',
                                                              'left_surface',
                                                              'right_surface'],
                                                 output_names=['out'],
                                                 function=utils.reformat_surfname),
                                    name='reformat_surface_name')
    reformat_surface_name.inputs.left_surface = white_surface_left
    reformat_surface_name.inputs.right_surface = white_surface_right
    reformat_surface_name.iterables = ('hemi', ['lh', 'rh'])

    mris_exp = pe.Node(niu.Function(input_names=['in_surface'],
                                    output_names=['out_surface'],
                                    function=utils.mris_expand),
                       name='mris_expand_white')

    surf_conversion = pe.MapNode(niu.Function(input_names=['in_surface',
                                                           'reg_file',
                                                           'gtmsegfile',
                                                           'subject_id',
                                                           'session_id',
                                                           'caps_dir'],
                                              output_names=['tval'],
                                              function=utils.surf2surf),
                                 name='surf_conversion',
                                 iterfield=['in_surface'])
    surf_conversion.inputs.subject_id = subject_id
    surf_conversion.inputs.session_id = session_id
    surf_conversion.inputs.caps_dir = caps_dir

    vol_on_surf = pe.MapNode(niu.Function(input_names=['volume',
                                                       'surface',
                                                       'subject_id',
                                                       'session_id',
                                                       'caps_dir',
                                                       'gtmsegfile'],
                                          output_names=['output'],
                                          function=utils.vol2surf),
                             name='vol_on_surf',
                             iterfield=['surface'])
    vol_on_surf.inputs.subject_id = subject_id
    vol_on_surf.inputs.session_id = session_id
    vol_on_surf.inputs.caps_dir = caps_dir

    normal_average = pe.Node(niu.Function(input_names=['in_surfaces'],
                                          output_names=['out_surface'],
                                          function=utils.weighted_mean),
                             name='normal_average')

    project_on_fsaverage = pe.Node(niu.Function(input_names=['projection',
                                                             'subject_id',
                                                             'caps_dir',
                                                             'session_id',
                                                             'fwhm'],
                                                output_names=['out_fsaverage'],
                                                function=utils.fsaverage_projection),
                                   name='project_on_fsaverage')
    project_on_fsaverage.iterables = ('fwhm', [0, 5, 10, 15, 20, 25])
    project_on_fsaverage.inputs.subject_id = subject_id
    project_on_fsaverage.inputs.session_id = session_id
    project_on_fsaverage.inputs.caps_dir = caps_dir

    extract_mid_surface = pe.Node(niu.Function(input_names=['in_surfaces'],
                                               output_names=['mid_surface'],
                                               function=utils.get_mid_surface),
                                  name='extract_mid_surface')

    psfreader = pe.Node(niu.Function(input_names=['json_path'],
                                     output_names=['in_plane', 'axial'],
                                     function=utils.readpsf),
                        name='psfreader')

    surface_atlas = {'destrieux': {'lh': destrieux_left,
                                   'rh': destrieux_right},
                     'desikan': {'lh': desikan_left,
                                 'rh': desikan_right}}

    gather_pet_projection = pe.JoinNode(niu.IdentityInterface(fields=['pet_projection_lh_rh']),
                                        name='gather_pet_projection_hemisphere',
                                        joinsource='reformat_surface_name',
                                        joinfield=['pet_projection_lh_rh'])

    atlas_tsv = pe.Node(
            niu.Function(input_names=['pet', 'atlas_files'],
                         output_names=['destrieux_tsv', 'desikan_tsv'],
                         function=utils.produce_tsv),
            name='atlas_tsv')
    atlas_tsv.inputs.atlas_files = surface_atlas

    # 2 creation of workflow : working dir, inputnode, outputnode and datasink

    wf = pe.Workflow(name=subject_id.replace('-', '_') + '_' + session_id.replace('-', '_'))
    wf.base_dir = working_directory_subjects

    inputnode = pe.Node(niu.IdentityInterface(fields=['orig_nu',
                                                      'pet',
                                                      'psf',
                                                      'white_surface_left',
                                                      'white_surface_right']),
                        name='inputnode',
                        mandatory_inputs=True)

    inputnode.inputs.orig_nu = orig_nu
    inputnode.inputs.pet = pet
    inputnode.inputs.psf = psf
    inputnode.inputs.white_surface_right = white_surface_right
    inputnode.inputs.white_surface_left = white_surface_left

    outputnode = pe.Node(niu.IdentityInterface(fields=['mid_surf',
                                                       'projection_native_subject',
                                                       'projection_fsaverage_smoothed',
                                                       'destrieux_tsv',
                                                       'desikan_tsv']),
                         name='outputnode',
                         mandatory_inputs=True)

    datasink = pe.Node(nio.DataSink(),
                       name='sinker')
    datasink.inputs.base_directory = os.path.join(caps_dir,
                                                  'subjects',
                                                  subject_id,
                                                  session_id,
                                                  'pet',
                                                  'surface')
    datasink.inputs.parameterization = True
    datasink.inputs.regexp_substitutions = [
        # Mid surface
        (r'(.*(sub-.*)\/(ses-.*)\/pet\/surface)\/midsurface\/.*_hemi_([a-z]+)(.*)$',
         r'\1/\2_\3_hemi-\4_midcorticalsurface'),
        # Projection in native space
        (r'(.*(sub-.*)\/(ses-.*)\/pet\/surface)\/projection_native\/.*_hemi_([a-z]+).*',
         r'\1/\2_\3_task-rest_acq-' + pet_tracer + r'_pet_space-native_suvr-' + utils.normalization_areas(pet_tracer) + r'_pvc-iy_hemi-\4_projection.mgh'),
        # Projection in fsaverage
        (r'(.*(sub-.*)\/(ses-.*)\/pet\/surface)\/projection_fsaverage\/.*_hemi_([a-z]+).*_fwhm_([0-9]+).*',
         r'\1/\2_\3_task-rest_acq-' + pet_tracer + r'_pet_space-fsaverage_suvr-' + utils.normalization_areas(pet_tracer) + r'_pvc-iy_hemi-\4_fwhm-\5_projection.mgh'),
        # TSV file for destrieux atlas
        (r'(.*(sub-.*)\/(ses-.*)\/pet\/surface)\/destrieux_tsv\/destrieux.tsv',
         r'\1/atlas_statistics/\2_\3_task-rest_acq-' + pet_tracer.lower() + '_pet_space-destrieux_pvc-iy_suvr-' + utils.normalization_areas(pet_tracer) + '_statistics.tsv'),
        # TSV file for desikan atlas
        (r'(.*(sub-.*)\/(ses-.*)\/pet\/surface)\/desikan_tsv\/desikan.tsv',
         r'\1/atlas_statistics/\2_\3_task-rest_acq-' + pet_tracer.lower() + '_pet_space-desikan_pvc-iy_suvr-' + utils.normalization_areas(pet_tracer) + '_statistics.tsv')
    ]

    # 3 Connecting the nodes

    # TODO(@arnaud.marcoux): Add titles for sections of connections.
    #   Could be useful to add sections title to group similar connections
    #   together.

    wf.connect([(inputnode, unzip_pet, [('pet', 'in_file')]),
                (unzip_pet, coreg, [('out_file', 'source')]),

                (inputnode, convert_mgh, [('orig_nu', 'in_file')]),
                (convert_mgh, unzip_orig_nu, [('out_file', 'in_file')]),

                (unzip_orig_nu, coreg, [('out_file', 'target')]),
                (coreg, removenan, [('coregistered_source', 'volname')]),
                (removenan, vol2vol, [('vol_wo_nan', 'source_file')]),

                (inputnode, tkregister, [('orig_nu', 'target_image')]),

                (unzip_orig_nu, normalize12, [('out_file', 'image_to_align')]),
                (unzip_mask, apply_inverse_deformation, [('out_file', 'img')]),
                (normalize12, apply_inverse_deformation, [('deformation_field', 'deformation_field')]),
                (unzip_orig_nu, apply_inverse_deformation, [('out_file', 'target')]),
                (apply_inverse_deformation, vol2vol_mask, [('freesurfer_space_eroded_mask', 'source_file')]),

                (gtmsegmentation, vol2vol_mask, [('gtmseg_file', 'target_file')]),
                (gtmsegmentation, tkregister, [('gtmseg_file', 'moving_image')]),
                (gtmsegmentation, convert_gtmseg, [('gtmseg_file', 'in_file')]),
                (gtmsegmentation, vol2vol, [('gtmseg_file', 'target_file')]),

                (vol2vol, pons_normalization, [('transformed_file', 'pet_path')]),
                (vol2vol_mask, pons_normalization, [('transformed_file', 'mask')]),

                (convert_gtmseg, labelconversion, [('out_file', 'gtmsegfile')]),
                (labelconversion, merge_volume, [('list_of_regions', 'in_files')]),

                (inputnode, psfreader, [('psf', 'json_path')]),

                (psfreader, pvc, [('in_plane', 'fwhm_x')]),
                (psfreader, pvc, [('in_plane', 'fwhm_y')]),
                (psfreader, pvc, [('axial', 'fwhm_z')]),
                (merge_volume, pvc, [('merged_file', 'mask_file')]),
                (pons_normalization, pvc, [('suvr', 'in_file')]),

                (reformat_surface_name, mris_exp, [('out', 'in_surface')]),
                (mris_exp, extract_mid_surface, [('out_surface', 'in_surfaces')]),

                (mris_exp, surf_conversion, [('out_surface', 'in_surface')]),
                (tkregister, surf_conversion, [('reg_file', 'reg_file')]),
                (gtmsegmentation, surf_conversion, [('gtmseg_file', 'gtmsegfile')]),

                (pvc, vol_on_surf, [('out_file', 'volume')]),
                (surf_conversion, vol_on_surf, [('tval', 'surface')]),
                (gtmsegmentation, vol_on_surf, [('gtmseg_file', 'gtmsegfile')]),

                (vol_on_surf, normal_average, [('output', 'in_surfaces')]),

                (normal_average, project_on_fsaverage, [('out_surface', 'projection')]),

                (normal_average, gather_pet_projection, [('out_surface', 'pet_projection_lh_rh')]),
                (gather_pet_projection, atlas_tsv, [('pet_projection_lh_rh', 'pet')]),
                (atlas_tsv, outputnode, [('destrieux_tsv', 'destrieux_tsv')]),
                (atlas_tsv, outputnode, [('desikan_tsv', 'desikan_tsv')]),
                (project_on_fsaverage, outputnode, [('out_fsaverage', 'projection_fsaverage_smoothed')]),
                (extract_mid_surface, outputnode, [('mid_surface', 'mid_surf')]),
                (normal_average, outputnode, [('out_surface', 'projection_native_subject')]),

                (outputnode, datasink, [('projection_fsaverage_smoothed', 'projection_fsaverage')]),
                (outputnode, datasink, [('mid_surf', 'midsurface')]),
                (outputnode, datasink, [('projection_native_subject', 'projection_native')]),
                (outputnode, datasink, [('destrieux_tsv', 'destrieux_tsv')]),
                (outputnode, datasink, [('desikan_tsv', 'desikan_tsv')])
                ])
    # wf.write_graph(graph2use='flat')
    wf.run()
    pass



# coding: utf8

from nipype import config

import clinica.pipelines.engine as cpe


# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
cfg = dict(execution={'parameterize_dirs': False})
config.update_config(cfg)


class PETVolume(cpe.Pipeline):
    """PETVolume - Volume-based processing of PET images using SPM.

    Returns:
        A clinica pipeline object containing the PETVolume pipeline.
    """
    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.group import check_group_label
        default_atlases = ['AAL2', 'LPBA40', 'Neuromorphometrics', 'AICHA', 'Hammers']

        if 'group_id' not in self.parameters.keys():
            raise KeyError('Missing compulsory group_id key in pipeline parameter.')
        if 'psf_tsv' not in self.parameters.keys():
            self.parameters['psf_tsv'] = None
        if 'pet_tracer' not in self.parameters.keys():
            self.parameters['pet_tracer'] = 'fdg'
        if 'mask_tissues' not in self.parameters.keys():
            self.parameters['mask_tissues'] = [1, 2, 3]
        if 'mask_threshold' not in self.parameters.keys():
            self.parameters['mask_threshold'] = 0.3
        if 'pvc_mask_tissues' not in self.parameters.keys():
            self.parameters['pvc_mask_tissues'] = [1, 2, 3]
        if 'smooth' not in self.parameters.keys():
            self.parameters['smooth'] = [8]
        if 'atlases' not in self.parameters.keys():
            self.parameters['atlases'] = default_atlases

        check_group_label(self.parameters['group_id'])

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """

        return ['pet_image',
                't1_image_native',
                'mask_tissues',
                'pvc_mask_tissues',
                'fwhm',
                'flow_fields',
                'dartel_template',
                'reference_mask']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """

        return ['pet_t1_native',
                'pet_mni',
                'pet_suvr',
                'binary_mask',
                'pet_suvr_masked',
                'pet_suvr_masked_smoothed',
                'pet_pvc',
                'pet_pvc_mni',
                'pet_pvc_suvr',
                'pet_pvc_suvr_masked',
                'pet_pvc_suvr_masked_smoothed',
                'atlas_statistics',
                'pvc_atlas_statistics'
                ]

    def build_input_node(self):
        """Build and connect an input node to the pipelines.
        """
        from colorama import Fore
        from os.path import join, split, realpath, exists
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from .pet_volume_utils import read_psf_information
        from clinica.utils.inputs import clinica_file_reader, clinica_group_reader, insensitive_glob
        from clinica.utils.input_files import (t1_volume_final_group_template,
                                               t1_volume_native_tpm,
                                               t1_volume_native_tpm_in_mni,
                                               t1_volume_deformation_to_template)
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.ux import print_groups_in_caps_directory
        from clinica.iotools.utils.data_handling import check_relative_volume_location_in_world_coordinate_system
        import clinica.utils.input_files as input_files

        # Check that group already exists
        if not exists(join(self.caps_directory, 'groups', 'group-' + self.parameters['group_id'])):
            print_groups_in_caps_directory(self.caps_directory)
            raise ClinicaException(
                '%sGroup %s does not exist. Did you run t1-volume or t1-volume-create-dartel pipeline?%s' %
                (Fore.RED, self.parameters['group_id'], Fore.RESET)
            )

        # Tissues DataGrabber
        # ====================
        all_errors = []

        # Grab reference mask
        if self.parameters['pet_tracer'] == 'fdg':
            reference_mask_template = 'region-pons_eroded-6mm_mask.nii*'
            self.parameters['suvr_region'] = 'pons'
            pet_file_to_grab = input_files.PET_FDG_NII
        elif self.parameters['pet_tracer'] == 'av45':
            reference_mask_template = 'region-cerebellumPons_eroded-6mm_mask.nii*'
            self.parameters['suvr_region'] = 'cerebellumPons'
            pet_file_to_grab = input_files.PET_AV45_NII
        else:
            raise NotImplementedError('Only "fdg" or "av45" tracers are currently accepted (given tracer "%s").' %
                                      self.parameters['pet_tracer'])
        reference_mask_path = join(split(realpath(__file__))[0], '../../resources/masks', reference_mask_template)
        reference_mask_file = insensitive_glob(reference_mask_path)
        if len(reference_mask_file) != 1:
            all_errors.append('An error happened while trying to get reference mask ' + str(reference_mask_path))
        else:
            reference_mask_file = reference_mask_file[0]

        # PET from BIDS directory
        try:
            pet_bids = clinica_file_reader(self.subjects,
                                           self.sessions,
                                           self.bids_directory,
                                           pet_file_to_grab)
        except ClinicaException as e:
            all_errors.append(e)

        # Native T1w-MRI
        try:
            t1w_bids = clinica_file_reader(self.subjects,
                                           self.sessions,
                                           self.bids_directory,
                                           input_files.T1W_NII)
        except ClinicaException as e:
            all_errors.append(e)

        # mask_tissues
        tissues_input = []
        for tissue_number in self.parameters['mask_tissues']:
            try:
                current_file = clinica_file_reader(self.subjects,
                                                   self.sessions,
                                                   self.caps_directory,
                                                   t1_volume_native_tpm_in_mni(tissue_number, False))
                tissues_input.append(current_file)
            except ClinicaException as e:
                all_errors.append(e)
        # Tissues_input has a length of len(self.parameters['mask_tissues']). Each of these elements has a size of
        # len(self.subjects). We want the opposite: a list of size len(self.subjects) whose elements have a size of
        # len(self.parameters['mask_tissues']. The trick is to iter on elements with zip(*my_list)
        tissues_input_final = []
        for subject_tissue_list in zip(*tissues_input):
            tissues_input_final.append(subject_tissue_list)
        tissues_input = tissues_input_final

        # Flowfields
        try:
            flowfields_caps = clinica_file_reader(self.subjects,
                                                  self.sessions,
                                                  self.caps_directory,
                                                  t1_volume_deformation_to_template(self.parameters['group_id']))
        except ClinicaException as e:
            all_errors.append(e)

        # Dartel Template
        try:
            final_template = clinica_group_reader(self.caps_directory,
                                                  t1_volume_final_group_template(self.parameters['group_id']))
        except ClinicaException as e:
            all_errors.append(e)

        if self.parameters['psf_tsv'] is not None:
            iterables_fwhm = read_psf_information(self.parameters['psf_tsv'], self.subjects, self.sessions)
            self.parameters['apply_pvc'] = True
        else:
            iterables_fwhm = [[]] * len(self.subjects)
            self.parameters['apply_pvc'] = False

        if self.parameters['apply_pvc']:
            # pvc tissues input
            pvc_tissues_input = []
            for tissue_number in self.parameters['pvc_mask_tissues']:
                try:
                    current_file = clinica_file_reader(self.subjects,
                                                       self.sessions,
                                                       self.caps_directory,
                                                       t1_volume_native_tpm(tissue_number))
                    pvc_tissues_input.append(current_file)
                except ClinicaException as e:
                    all_errors.append(e)

            if len(all_errors) == 0:
                pvc_tissues_input_final = []
                for subject_tissue_list in zip(*pvc_tissues_input):
                    pvc_tissues_input_final.append(subject_tissue_list)
                pvc_tissues_input = pvc_tissues_input_final
        else:
            pvc_tissues_input = []

        if len(all_errors) > 0:
            error_message = 'Clinica faced error(s) while trying to read files in your CAPS/BIDS directories.\n'
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaException(error_message)

        check_relative_volume_location_in_world_coordinate_system('T1w-MRI', t1w_bids,
                                                                  self.parameters['pet_tracer'] + ' PET', pet_bids,
                                                                  self.bids_directory,
                                                                  self.parameters['pet_tracer'])

        read_input_node = npe.Node(name="LoadingCLIArguments",
                                   interface=nutil.IdentityInterface(
                                       fields=self.get_input_fields(),
                                       mandatory_inputs=True),
                                   iterables=[('pet_image', pet_bids),
                                              ('t1_image_native', t1w_bids),
                                              ('mask_tissues', tissues_input),
                                              ('fwhm', iterables_fwhm),
                                              ('flow_fields', flowfields_caps),
                                              ('pvc_mask_tissues', pvc_tissues_input)],
                                   synchronize=True)

        read_input_node.inputs.reference_mask = reference_mask_file
        read_input_node.inputs.dartel_template = final_template

        self.connect([(read_input_node, self.input_node, [('pet_image', 'pet_image')]),
                      (read_input_node, self.input_node, [('t1_image_native', 't1_image_native')]),
                      (read_input_node, self.input_node, [('mask_tissues', 'mask_tissues')]),
                      (read_input_node, self.input_node, [('flow_fields', 'flow_fields')]),
                      (read_input_node, self.input_node, [('dartel_template', 'dartel_template')]),
                      (read_input_node, self.input_node, [('reference_mask', 'reference_mask')]),
                      (read_input_node, self.input_node, [('fwhm', 'fwhm')]),
                      (read_input_node, self.input_node, [('pvc_mask_tissues', 'pvc_mask_tissues')])])

    def build_output_node(self):
        """Build and connect an output node to the pipelines.
        """

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio
        from clinica.utils.filemanip import zip_nii
        from clinica.utils.nipype import fix_join
        import re
        import clinica.pipelines.pet_volume.pet_volume_utils as utils

        # Find container path from pet filename
        # =====================================
        container_path = npe.Node(nutil.Function(input_names=['pet_filename'],
                                                 output_names=['container'],
                                                 function=utils.pet_container_from_filename),
                                  name='container_path')
        container_path.inputs.threshold = self.parameters['mask_threshold']

        # Writing all images into CAPS
        # ============================
        write_images_node = npe.Node(name='write_caps_node',
                                     interface=nio.DataSink())
        write_images_node.inputs.base_directory = self.caps_directory
        write_images_node.inputs.parameterization = False
        write_images_node.inputs.regexp_substitutions = [
            (r'(.*/)pet_t1_native/r(sub-.*)(\.nii(\.gz)?)$', r'\1\2_space-T1w_pet\3'),
            (r'(.*/)pet_pvc/pvc-rbv_r(sub-.*)(\.nii(\.gz)?)$', r'\1\2_space-T1w_pvc-rbv_pet\3'),
            (r'(.*/)pet_mni/wr(sub-.*)(\.nii(\.gz)?)$', r'\1\2_space-Ixi549Space_pet\3'),
            (r'(.*/)pet_pvc_mni/wpvc-rbv_r(sub-.*)(\.nii(\.gz)?)$', r'\1\2_space-Ixi549Space_pvc-rbv_pet\3'),
            (r'(.*/)pet_suvr/suvr_wr(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_space-Ixi549Space_suvr-' + re.escape(self.parameters['suvr_region']) + r'_pet\3'),
            (r'(.*/)pet_pvc_suvr/suvr_wpvc-rbv_r(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_space-Ixi549Space_pvc-rbv_suvr-' + re.escape(self.parameters['suvr_region']) + r'_pet\3'),
            (r'(.*/)pet_suvr_masked/masked_suvr_wr(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_space-Ixi549Space_suvr-' + re.escape(self.parameters['suvr_region']) + r'_mask-brain_pet\3'),
            (r'(.*/)pet_pvc_suvr_masked/masked_suvr_wpvc-rbv_r(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_space-Ixi549Space_pvc-rbv_suvr-' + re.escape(self.parameters['suvr_region']) + r'_mask-brain_pet\3'),
            (r'(.*/)pet_suvr_masked_smoothed/(fwhm-[0-9]+mm)_masked_suvr_wr(sub-.*)(\.nii(\.gz)?)$',
             r'\1\3_space-Ixi549Space_suvr-' + re.escape(self.parameters['suvr_region']) + r'_mask-brain_\2_pet\4'),
            (r'(.*/)pet_pvc_suvr_masked_smoothed/(fwhm-[0-9]+mm)_masked_suvr_wpvc-rbv_r(sub-.*)(\.nii(\.gz)?)$',
             r'\1\3_space-Ixi549Space_pvc-rbv_suvr-' + re.escape(self.parameters['suvr_region']) + r'_mask-brain_\2_pet\4'),
            (r'(.*/)binary_mask/(sub-.*_T1w_).*(space-[a-zA-Z0-9]+).*(_brainmask\.nii(\.gz)?)$', r'\1\2\3\4')
        ]

        # Writing atlas statistics into CAPS
        # ==================================
        write_atlas_node = npe.Node(name='write_atlas_node',
                                    interface=nio.DataSink())
        write_atlas_node.inputs.base_directory = self.caps_directory
        write_atlas_node.inputs.parameterization = False
        write_atlas_node.inputs.regexp_substitutions = [
            (r'(.*/atlas_statistics/)suvr_wr(sub-.*)(_statistics\.tsv)$', r'\1\2' + r'_suvr-' +
             re.escape(self.parameters['suvr_region']) + r'\3'),
            (r'(.*/)pvc_(atlas_statistics/)suvr_wpvc-rbv_r(sub-.*)(_statistics\.tsv)$', r'\1\2\3' + r'_pvc-rbv_suvr-' +
             re.escape(self.parameters['suvr_region']) + r'\4')
        ]

        self.connect([(self.input_node, container_path, [('pet_image', 'pet_filename')]),
                      (container_path, write_images_node, [(('container', fix_join, 'group-' + self.parameters['group_id']),
                                                            'container')]),
                      (self.output_node, write_images_node, [(('pet_t1_native', zip_nii, True), 'pet_t1_native'),
                                                             (('pet_mni', zip_nii, True), 'pet_mni'),
                                                             (('pet_suvr', zip_nii, True), 'pet_suvr'),
                                                             (('binary_mask', zip_nii, True), 'binary_mask'),
                                                             (('pet_suvr_masked', zip_nii, True), 'pet_suvr_masked'),
                                                             (('pet_suvr_masked_smoothed', zip_nii, True),
                                                              'pet_suvr_masked_smoothed'),
                                                             (('pet_pvc', zip_nii, True), 'pet_pvc'),
                                                             (('pet_pvc_mni', zip_nii, True), 'pet_pvc_mni'),
                                                             (('pet_pvc_suvr', zip_nii, True), 'pet_pvc_suvr'),
                                                             (('pet_pvc_suvr_masked', zip_nii, True),
                                                              'pet_pvc_suvr_masked'),
                                                             (('pet_pvc_suvr_masked_smoothed', zip_nii, True),
                                                              'pet_pvc_suvr_masked_smoothed')]),
                      (container_path, write_atlas_node, [(('container', fix_join, 'group-' + self.parameters['group_id']),
                                                           'container')]),
                      (self.output_node, write_atlas_node, [('atlas_statistics', 'atlas_statistics'),
                                                            ('pvc_atlas_statistics', 'pvc_atlas_statistics')])
                      ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipelines.
        """

        import nipype.interfaces.spm as spm
        import nipype.interfaces.spm.utils as spmutils
        from nipype.interfaces.petpvc import PETPVC
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.utils.filemanip import unzip_nii
        import clinica.pipelines.pet_volume.pet_volume_utils as utils

        import os

        if 'SPMSTANDALONE_HOME' in os.environ:
            if 'MCR_HOME' in os.environ:
                matlab_cmd = (
                        os.path.join(
                            os.environ['SPMSTANDALONE_HOME'], 'run_spm12.sh')
                        + ' ' + os.environ['MCR_HOME']
                        + ' script')
                spm.SPMCommand.set_mlab_paths(matlab_cmd=matlab_cmd, use_mcr=True)

        # Unzipping
        # ==================
        unzip_pet_image = npe.Node(nutil.Function(input_names=['in_file'],
                                                  output_names=['out_file'],
                                                  function=unzip_nii),
                                   name='unzip_pet_image')

        unzip_t1_image_native = npe.Node(nutil.Function(input_names=['in_file'],
                                                        output_names=['out_file'],
                                                        function=unzip_nii),
                                         name='unzip_t1_image_native')

        unzip_flow_fields = npe.Node(nutil.Function(input_names=['in_file'],
                                                    output_names=['out_file'],
                                                    function=unzip_nii),
                                     name='unzip_flow_fields')

        unzip_dartel_template = npe.Node(nutil.Function(input_names=['in_file'],
                                                        output_names=['out_file'],
                                                        function=unzip_nii),
                                         name='unzip_dartel_template')

        unzip_reference_mask = npe.Node(nutil.Function(input_names=['in_file'],
                                                       output_names=['out_file'],
                                                       function=unzip_nii),
                                        name='unzip_reference_mask')

        unzip_mask_tissues = npe.MapNode(nutil.Function(input_names=['in_file'],
                                                        output_names=['out_file'],
                                                        function=unzip_nii),
                                         name='unzip_mask_tissues',
                                         iterfield=['in_file'])

        # Coregister PET into T1 native space
        # ===================================
        coreg_pet_t1 = npe.Node(spm.Coregister(), name='coreg_pet_t1')

        # Spatially normalize PET into MNI
        # ================================
        dartel_mni_reg = npe.Node(spm.DARTELNorm2MNI(), name='dartel_mni_reg')
        dartel_mni_reg.inputs.modulate = False
        dartel_mni_reg.inputs.fwhm = 0

        # Reslice reference region mask into PET
        # ======================================
        reslice = npe.Node(spmutils.Reslice(), name='reslice')

        # Normalize PET values according to reference region
        # ==================================================
        norm_to_ref = npe.Node(nutil.Function(input_names=['pet_image', 'region_mask'],
                                              output_names=['suvr_pet_path'],
                                              function=utils.normalize_to_reference),
                               name='norm_to_ref')

        # Create binary mask from segmented tissues
        # =========================================
        binary_mask = npe.Node(nutil.Function(input_names=['tissues', 'threshold'],
                                              output_names=['out_mask'],
                                              function=utils.create_binary_mask),
                               name='binary_mask')
        binary_mask.inputs.threshold = self.parameters['mask_threshold']

        # Mask PET image
        # ==============
        apply_mask = npe.Node(nutil.Function(input_names=['image', 'binary_mask'],
                                             output_names=['masked_image_path'],
                                             function=utils.apply_binary_mask),
                              name='apply_mask')

        # Smoothing
        # =========
        if self.parameters['smooth'] is not None and len(self.parameters['smooth']) > 0:
            smoothing_node = npe.MapNode(spm.Smooth(),
                                         name='smoothing_node',
                                         iterfield=['fwhm', 'out_prefix'])
            smoothing_node.inputs.fwhm = [[x, x, x] for x in self.parameters['smooth']]
            smoothing_node.inputs.out_prefix = ['fwhm-' + str(x) + 'mm_' for x in self.parameters['smooth']]
            self.connect([
                (apply_mask, smoothing_node, [('masked_image_path', 'in_files')]),
                (smoothing_node, self.output_node, [('smoothed_files', 'pet_suvr_masked_smoothed')])
            ])
        else:
            self.output_node.inputs.pet_suvr_masked_smoothed = [[]]

        # Atlas Statistics
        # ================
        atlas_stats_node = npe.MapNode(nutil.Function(input_names=['in_image',
                                                                   'in_atlas_list'],
                                                      output_names=['atlas_statistics'],
                                                      function=utils.atlas_statistics),
                                       name='atlas_stats_node',
                                       iterfield=['in_image'])
        atlas_stats_node.inputs.in_atlas_list = self.parameters['atlases']

        # Connection
        # ==========
        self.connect([(self.input_node, unzip_pet_image, [('pet_image', 'in_file')]),
                      (self.input_node, unzip_t1_image_native, [('t1_image_native', 'in_file')]),
                      (self.input_node, unzip_flow_fields, [('flow_fields', 'in_file')]),
                      (self.input_node, unzip_dartel_template, [('dartel_template', 'in_file')]),
                      (self.input_node, unzip_reference_mask, [('reference_mask', 'in_file')]),
                      (self.input_node, unzip_mask_tissues, [('mask_tissues', 'in_file')]),

                      (unzip_pet_image, coreg_pet_t1, [('out_file', 'source')]),
                      (unzip_t1_image_native, coreg_pet_t1, [('out_file', 'target')]),
                      (unzip_flow_fields, dartel_mni_reg, [('out_file', 'flowfield_files')]),
                      (unzip_dartel_template, dartel_mni_reg, [('out_file', 'template_file')]),
                      (unzip_reference_mask, reslice, [('out_file', 'in_file')]),
                      (unzip_mask_tissues, binary_mask, [('out_file', 'tissues')]),

                      (coreg_pet_t1, dartel_mni_reg, [('coregistered_source', 'apply_to_files')]),
                      (dartel_mni_reg, reslice, [('normalized_files', 'space_defining')]),
                      (dartel_mni_reg, norm_to_ref, [('normalized_files', 'pet_image')]),
                      (reslice, norm_to_ref, [('out_file', 'region_mask')]),
                      (norm_to_ref, apply_mask, [('suvr_pet_path', 'image')]),
                      (binary_mask, apply_mask, [('out_mask', 'binary_mask')]),
                      (norm_to_ref, atlas_stats_node, [('suvr_pet_path', 'in_image')]),

                      (coreg_pet_t1, self.output_node, [('coregistered_source', 'pet_t1_native')]),
                      (dartel_mni_reg, self.output_node, [('normalized_files', 'pet_mni')]),
                      (norm_to_ref, self.output_node, [('suvr_pet_path', 'pet_suvr')]),
                      (binary_mask, self.output_node, [('out_mask', 'binary_mask')]),
                      (apply_mask, self.output_node, [('masked_image_path', 'pet_suvr_masked')]),
                      (atlas_stats_node, self.output_node, [('atlas_statistics', 'atlas_statistics')])
                      ])

        # PVC
        # ==========
        if self.parameters['apply_pvc']:
            # Unzipping
            # =========
            unzip_pvc_mask_tissues = npe.MapNode(nutil.Function(input_names=['in_file'],
                                                                output_names=['out_file'],
                                                                function=unzip_nii),
                                                 name='unzip_pvc_mask_tissues',
                                                 iterfield=['in_file'])

            # Creating Mask to use in PVC
            # ===========================
            pvc_mask = npe.Node(nutil.Function(input_names=['tissues'],
                                               output_names=['out_mask'],
                                               function=utils.create_pvc_mask),
                                name='pvc_mask')
            # PET PVC
            # =======
            petpvc = npe.Node(PETPVC(), name='pvc')
            petpvc.inputs.pvc = 'RBV'
            petpvc.inputs.out_file = 'pvc.nii'

            # Spatially normalize PET into MNI
            # ================================
            dartel_mni_reg_pvc = npe.Node(spm.DARTELNorm2MNI(), name='dartel_mni_reg_pvc')
            dartel_mni_reg_pvc.inputs.modulate = False
            dartel_mni_reg_pvc.inputs.fwhm = 0

            # Reslice reference region mask into PET
            # ======================================
            reslice_pvc = npe.Node(spmutils.Reslice(), name='reslice_pvc')

            # Normalize PET values according to reference region
            # ==================================================
            norm_to_ref_pvc = npe.Node(nutil.Function(input_names=['pet_image', 'region_mask'],
                                                      output_names=['suvr_pet_path'],
                                                      function=utils.normalize_to_reference),
                                       name='norm_to_ref_pvc')

            # Mask PET image
            # ==============
            apply_mask_pvc = npe.Node(nutil.Function(input_names=['image', 'binary_mask'],
                                                     output_names=['masked_image_path'],
                                                     function=utils.apply_binary_mask),
                                      name='apply_mask_pvc')
            # Smoothing
            # =========
            if self.parameters['smooth'] is not None and len(self.parameters['smooth']) > 0:
                smoothing_pvc = npe.MapNode(spm.Smooth(),
                                            name='smoothing_pvc',
                                            iterfield=['fwhm', 'out_prefix'])
                smoothing_pvc.inputs.fwhm = [[x, x, x] for x in self.parameters['smooth']]
                smoothing_pvc.inputs.out_prefix = ['fwhm-' + str(x) + 'mm_' for x in self.parameters['smooth']]
                self.connect([
                    (apply_mask_pvc, smoothing_pvc, [('masked_image_path', 'in_files')]),
                    (smoothing_pvc, self.output_node, [('smoothed_files', 'pet_pvc_suvr_masked_smoothed')])
                ])
            else:
                self.output_node.inputs.pet_pvc_suvr_masked_smoothed = [[]]
            # Atlas Statistics
            # ================
            atlas_stats_pvc = npe.MapNode(nutil.Function(input_names=['in_image',
                                                                      'in_atlas_list'],
                                                         output_names=['atlas_statistics'],
                                                         function=utils.atlas_statistics),
                                          name='atlas_stats_pvc',
                                          iterfield=['in_image'])
            atlas_stats_pvc.inputs.in_atlas_list = self.parameters['atlases']

            # Connection
            # ==========
            self.connect([(self.input_node, unzip_pvc_mask_tissues, [('pvc_mask_tissues', 'in_file')]),
                          (unzip_pvc_mask_tissues, pvc_mask, [('out_file', 'tissues')]),
                          (unzip_flow_fields, dartel_mni_reg_pvc, [('out_file', 'flowfield_files')]),
                          (unzip_dartel_template, dartel_mni_reg_pvc, [('out_file', 'template_file')]),
                          (unzip_reference_mask, reslice_pvc, [('out_file', 'in_file')]),
                          (coreg_pet_t1, petpvc, [('coregistered_source', 'in_file'),
                                                  (('coregistered_source', utils.pet_pvc_name, 'RBV'), 'out_file')]),
                          (pvc_mask, petpvc, [('out_mask', 'mask_file')]),
                          (self.input_node, petpvc, [(('fwhm', utils.get_from_list, 0), 'fwhm_x'),
                                                     (('fwhm', utils.get_from_list, 1), 'fwhm_y'),
                                                     (('fwhm', utils.get_from_list, 2), 'fwhm_z')]),
                          (petpvc, dartel_mni_reg_pvc, [('out_file', 'apply_to_files')]),
                          (dartel_mni_reg_pvc, reslice_pvc, [('normalized_files', 'space_defining')]),
                          (dartel_mni_reg_pvc, norm_to_ref_pvc, [('normalized_files', 'pet_image')]),
                          (reslice_pvc, norm_to_ref_pvc, [('out_file', 'region_mask')]),

                          (norm_to_ref_pvc, apply_mask_pvc, [('suvr_pet_path', 'image')]),
                          (binary_mask, apply_mask_pvc, [('out_mask', 'binary_mask')]),
                          (norm_to_ref_pvc, atlas_stats_pvc, [('suvr_pet_path', 'in_image')]),

                          (petpvc, self.output_node, [('out_file', 'pet_pvc')]),
                          (dartel_mni_reg_pvc, self.output_node, [('normalized_files', 'pet_pvc_mni')]),
                          (norm_to_ref_pvc, self.output_node, [('suvr_pet_path', 'pet_pvc_suvr')]),
                          (apply_mask_pvc, self.output_node, [('masked_image_path', 'pet_pvc_suvr_masked')]),
                          (atlas_stats_pvc, self.output_node, [('atlas_statistics', 'pvc_atlas_statistics')])
                          ])
        else:
            self.output_node.inputs.pet_pvc = [[]]
            self.output_node.inputs.pet_pvc_mni = [[]]
            self.output_node.inputs.pet_pvc_suvr = [[]]
            self.output_node.inputs.pet_pvc_suvr_masked = [[]]
            self.output_node.inputs.pvc_atlas_statistics = [[]]
            self.output_node.inputs.pet_pvc_suvr_masked_smoothed = [[]]



# coding: utf8


def zip_nii(in_file, same_dir=False):
    from os import getcwd
    from os.path import abspath, join
    import gzip
    import shutil
    from nipype.utils.filemanip import split_filename
    from traits.trait_base import _Undefined

    if (in_file is None) or isinstance(in_file, _Undefined):
        return None

    if not isinstance(in_file, str):  # type(in_file) is list:
        return [zip_nii(f, same_dir) for f in in_file]

    orig_dir, base, ext = split_filename(str(in_file))

    # Already compressed
    if ext[-3:].lower() == ".gz":
        return in_file
    # Not compressed

    if same_dir:
        out_file = abspath(join(orig_dir, base + ext + '.gz'))
    else:
        out_file = abspath(join(getcwd(), base + ext + '.gz'))

    with open(in_file, 'rb') as f_in, gzip.open(out_file, 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)

    return out_file


def unzip_nii(in_file):
    from nipype.utils.filemanip import split_filename
    from nipype.algorithms.misc import Gunzip
    from traits.trait_base import _Undefined

    if (in_file is None) or isinstance(in_file, _Undefined):
        return None

    if not isinstance(in_file, str):  # type(in_file) is list:
        return [unzip_nii(f) for f in in_file]

    _, base, ext = split_filename(in_file)

    # Not compressed
    if ext[-3:].lower() != ".gz":
        return in_file
    # Compressed
    gunzip = Gunzip(in_file=in_file)
    gunzip.run()
    return gunzip.aggregate_outputs().out_file


def save_participants_sessions(participant_ids, session_ids, out_folder, out_file=None):
    """
    Save <participant_ids> <session_ids> in <out_folder>/<out_file> TSV file.
    """
    import os
    import errno
    import pandas
    from clinica.utils.stream import cprint

    assert(len(participant_ids) == len(session_ids))

    try:
        os.makedirs(out_folder)
    except OSError as e:
        if e.errno != errno.EEXIST:  # EEXIST: folder already exists
            raise e

    if out_file:
        tsv_file = os.path.join(out_folder, out_file)
    else:
        tsv_file = os.path.join(out_folder, 'participants.tsv')

    try:
        data = pandas.DataFrame({
            'participant_id': participant_ids,
            'session_id': session_ids,
        })
        data.to_csv(tsv_file, sep='\t', index=False, encoding='utf-8')
    except Exception as e:
        cprint("Impossible to save %s with pandas" % out_file)
        raise e


def get_subject_id(bids_or_caps_file):
    """
    Extracts "sub-<participant_id>_ses-<session_label>" from BIDS or CAPS file
    """
    import re

    m = re.search(r'(sub-[a-zA-Z0-9]+)/(ses-[a-zA-Z0-9]+)', bids_or_caps_file)

    if m is None:
        raise ValueError(
            'Input filename is not in a BIDS or CAPS compliant format.'
            ' It does not contain the subject and session information.')

    subject_id = m.group(1) + '_' + m.group(2)

    return subject_id


def extract_image_ids(bids_or_caps_files):
    """Extract image IDs (e.g. ['sub-CLNC01_ses-M00', 'sub-CLNC01_ses-M18']  from `bids_or_caps_files`."""
    import re
    id_bids_or_caps_files = [re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)', file).group()
                             for file in bids_or_caps_files]
    return id_bids_or_caps_files


def extract_subjects_sessions_from_filename(bids_or_caps_files):
    """Extract subjects/sessions (e.g. ['sub-CLNC01', 'sub-CLNC01']/['ses-M00', 'ses-M18'] from `bids_or_caps_files`."""
    id_bids_or_caps_files = extract_image_ids(bids_or_caps_files)
    split = [image_id.split('_')
             for image_id in id_bids_or_caps_files]
    subject_ids = [p_id[0] for p_id in split]
    session_ids = [s_id[1] for s_id in split]
    return subject_ids, session_ids


def extract_crash_files_from_log_file(filename):
    """Extract crash files (*.pklz) from `filename`.
    """
    import os
    import re

    assert(os.path.isfile(filename)),\
        'extract_crash_files_from_log_file: filename parameter is not a file (%s)' % filename

    log_file = open(filename, "r")
    crash_files = []
    for line in log_file:
        if re.match("(.*)crashfile:(.*)", line):
            crash_files.append(line.replace('\t crashfile:', '').replace('\n', ''))

    return crash_files


def read_participant_tsv(tsv_file):
    """Extract participant IDs and session IDs from TSV file.

    Raise:
        ClinicaException if tsv_file is not a file
        ClinicaException if participant_id or session_id column is missing from TSV file
    """
    import os
    import pandas as pd
    from colorama import Fore
    from clinica.utils.exceptions import ClinicaException

    if not os.path.isfile(tsv_file):
        raise ClinicaException(
            "\n%s[Error] The TSV file you gave is not a file.%s\n"
            "\n%sError explanations:%s\n"
            " - Clinica expected the following path to be a file: %s%s%s\n"
            " - If you gave relative path, did you run Clinica on the good folder?" %
            (Fore.RED, Fore.RESET,
             Fore.YELLOW, Fore.RESET,
             Fore.BLUE, tsv_file, Fore.RESET)
        )
    ss_df = pd.io.parsers.read_csv(tsv_file, sep='\t')
    if 'participant_id' not in list(ss_df.columns.values):
        raise ClinicaException(
            "\n%s[Error] The TSV file does not contain participant_id column (path: %s)%s" %
            (Fore.RED, tsv_file, Fore.RESET)
        )
    if 'session_id' not in list(ss_df.columns.values):
        raise ClinicaException(
            "\n%s[Error] The TSV file does not contain session_id column (path: %s)%s" %
            (Fore.RED, tsv_file, Fore.RESET)
        )
    participants = list(ss_df.participant_id)
    sessions = list(ss_df.session_id)

    # Remove potential whitespace in participant_id or session_id
    return [sub.strip(' ') for sub in participants], [ses.strip(' ') for ses in sessions]



# coding: utf-8

"""
 Module for converting fMRI of ADNI
"""

__author__ = "Jorge Samper-Gonzalez and Sabrina Fontanella"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_fmri(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Convert fMR images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of fMRI images. Output will be stored in ' + path.join(dest_dir, 'conversion_info') + '.')
    images = compute_fmri_path(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of fMRI images found. Exporting images into BIDS ...')
    fmri_paths_to_bids(dest_dir, images)
    cprint(Fore.GREEN + 'fMRI conversion done.' + Fore.RESET)


def compute_fmri_path(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute the paths to fMR images

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns: pandas Dataframe containing the path for each fmri

    """

    import operator
    from functools import reduce
    from os import path, mkdir
    import pandas as pd
    from clinica.iotools.converters.adni_to_bids.adni_utils import find_image_path, visits_to_timepoints_mrilist

    fmri_col = ['Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date',
                'Study_ID', 'Field_Strength', 'Series_ID', 'Image_ID']
    fmri_df = pd.DataFrame(columns=fmri_col)

    # Loading needed .csv files
    adni_merge = pd.read_csv(path.join(csv_dir, 'ADNIMERGE.csv'), sep=',', low_memory=False)

    mayo_mri_qc = pd.read_csv(path.join(csv_dir, 'MAYOADIRL_MRI_IMAGEQC_12_08_15.csv'), sep=',', low_memory=False)
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == 'fMRI']
    mayo_mri_qc.columns = [x.upper() for x in mayo_mri_qc.columns]

    mayo_mri_qc3 = pd.read_csv(path.join(csv_dir, 'MAYOADIRL_MRI_QUALITY_ADNI3.csv'), sep=',', low_memory=False)
    mayo_mri_qc3 = mayo_mri_qc3[mayo_mri_qc3.SERIES_TYPE == 'EPB']

    # Concatenating visits in both QC files
    mayo_mri_qc = pd.concat([mayo_mri_qc, mayo_mri_qc3], axis=0, ignore_index=True, sort=False)

    mri_list = pd.read_csv(path.join(csv_dir, 'MRILIST.csv'), sep=',', low_memory=False)

    # Selecting only fMRI images that are not Multiband
    mri_list = mri_list[mri_list.SEQUENCE.str.contains('MRI')]  # 'MRI' includes all fMRI and fMRI scans, but not others
    unwanted_sequences = ['MB']
    mri_list = mri_list[mri_list.SEQUENCE.map(lambda x: not any(subs in x for subs in unwanted_sequences))]

    # We will convert the images for each subject in the subject list
    for subj in subjs_list:

        # Filter ADNIMERGE, MRI_LIST and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        adnimerge_subj = adnimerge_subj.sort_values('EXAMDATE')

        mri_list_subj = mri_list[mri_list.SUBJECT == subj]
        mri_list_subj = mri_list_subj.sort_values('SCANDATE')

        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints_mrilist(subj, mri_list_subj, adnimerge_subj, "fMRI")

        for visit_info in visits.keys():
            timepoint = visit_info[0]
            visit_str = visits[visit_info]

            visit_mri_list = mri_list_subj[mri_list_subj.VISIT == visit_str]
            image = fmri_image(subj, timepoint, visits[visit_info], visit_mri_list, mayo_mri_qc_subj)

            if image is not None:
                row_to_append = pd.DataFrame(image, index=['i', ])
                # TODO Replace iteratively appending by pandas.concat
                fmri_df = fmri_df.append(row_to_append, ignore_index=True, sort=False)

    # Exceptions
    # ==========
    conversion_errors = [('006_S_4485', 'm84')]

    error_indices = []
    for conv_error in conversion_errors:
        error_indices.append((fmri_df.Subject_ID == conv_error[0])
                             & (fmri_df.VISCODE == conv_error[1]))

    if error_indices:
        indices_to_remove = fmri_df.index[reduce(operator.or_, error_indices, False)]
        fmri_df.drop(indices_to_remove, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(fmri_df, source_dir, 'fMRI', 'S', 'Series_ID')

    fmri_tsv_path = path.join(dest_dir, 'conversion_info')
    if not path.exists(fmri_tsv_path):
        mkdir(fmri_tsv_path)
    images.to_csv(path.join(fmri_tsv_path, 'fmri_paths.tsv'), sep='\t', index=False)

    return images


def fmri_paths_to_bids(dest_dir, fmri_paths, mod_to_update=False):
    """
    Convert the fmri extracted from the fmri_paths to BIDS

    Args:
        dest_dir: path to the input directory
        fmri_paths: path to the BIDS directory
        mod_to_update:  if True overwrite (or create if is missing) all the existing fmri

    """
    from multiprocessing import cpu_count, Pool, Value
    from functools import partial

    counter = None

    def init(args):
        # store the counter for later use
        global counter
        counter = args

    subjs_list = fmri_paths['Subject_ID'].drop_duplicates().values

    counter = Value('i', 0)
    partial_generate_subject_files = partial(generate_subject_files,
                                             fmri_paths=fmri_paths,
                                             dest_dir=dest_dir,
                                             mod_to_update=mod_to_update)
    poolrunner = Pool(cpu_count(), initializer=init, initargs=(counter,))
    poolrunner.map(partial_generate_subject_files, subjs_list)
    del counter


def generate_subject_files(subj, fmri_paths, dest_dir, mod_to_update):
    import clinica.iotools.converters.adni_to_bids.adni_utils as adni_utils
    import clinica.iotools.bids_utils as bids_utils
    from clinica.utils.stream import cprint
    import os
    from os import path
    from glob import glob

    sess_list = fmri_paths[(fmri_paths['Subject_ID'] == subj)]['VISCODE'].values
    alpha_id = adni_utils.remove_space_and_symbols(subj)
    bids_id = 'sub-ADNI' + alpha_id

    # For each session available, create the folder if doesn't exist and convert the files
    for ses in sess_list:
        with counter.get_lock():
            counter.value += 1
        cprint('[fMRI] Processing subject ' + str(subj)
               + ' - session ' + ses + ', ' + str(counter.value)
               + ' / ' + str(len(fmri_paths)))
        ses_bids = adni_utils.viscode_to_session(ses)
        bids_ses_id = 'ses-' + ses_bids
        bids_file_name = bids_id + '_ses-' + ses_bids
        ses_path = path.join(dest_dir, bids_id, bids_ses_id)

        # If the fmri already exist
        existing_fmri = glob(path.join(ses_path, 'func', '*_bold*'))
        # if mod_to_add:
        #     if len(existing_fmri) > 0:
        #         print 'Fmri already existing. Skipped.'
        #         continue

        if mod_to_update and len(existing_fmri) > 0:
            # print 'Removing the old fmri folder...'
            os.remove(existing_fmri[0])

        if not os.path.exists(ses_path):
            if not os.path.exists(path.join(dest_dir, bids_id)):
                os.mkdir(path.join(dest_dir, bids_id))
            os.mkdir(path.join(dest_dir, bids_id, bids_ses_id))

        fmri_info = fmri_paths[(fmri_paths['Subject_ID'] == subj) & (fmri_paths['VISCODE'] == ses)]
        if not fmri_info['Path'].values[0] == '':
            if type(fmri_info['Path'].values[0]) != float:
                if not os.path.exists(path.join(ses_path, 'func')):
                    os.mkdir(path.join(ses_path, 'func'))
                fmri_path = fmri_info['Path'].values[0]
                dcm_to_convert = adni_utils.check_two_dcm_folder(fmri_path,
                                                                 dest_dir,
                                                                 fmri_info['Image_ID'].values[0])
                if not os.path.isfile(os.path.join(ses_path, 'func', bids_file_name + '_task-rest_bold.nii.gz')):
                    bids_utils.convert_fmri(dcm_to_convert, path.join(ses_path, 'func'), bids_file_name)
                else:
                    cprint("Images already converted")

                # Delete the temporary folder used for copying fmri with 2 subjects inside the DICOM folder
                adni_utils.remove_tmp_dmc_folder(dest_dir, fmri_info['Image_ID'].values[0])


def fmri_image(subject_id, timepoint, visit_str, visit_mri_list, mri_qc_subj):
    """

    Args:
        subject_id:
        timepoint:
        visit_str:
        visit_mri_list:
        mri_qc_subj:

    Returns:

    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars, select_image_qc

    mri_qc_subj.columns = [x.lower() for x in mri_qc_subj.columns]
    sel_image = select_image_qc(list(visit_mri_list.IMAGEUID), mri_qc_subj)
    if sel_image is None:
        return None

    sel_scan = visit_mri_list[visit_mri_list.IMAGEUID == sel_image].iloc[0]

    image_dict = {'Subject_ID': subject_id,
                  'VISCODE': timepoint,
                  'Visit': visit_str,
                  'Sequence': replace_sequence_chars(sel_scan.SEQUENCE),
                  'Scan_Date': sel_scan['SCANDATE'],
                  'Study_ID': str(int(sel_scan.STUDYID)),
                  'Series_ID': str(int(sel_scan.SERIESID)),
                  'Image_ID': str(int(sel_scan.IMAGEUID)),
                  'Field_Strength': sel_scan.MAGSTRENGTH}

    return image_dict


def check_exceptions(bids_dir):

    from os import path
    import pandas as pd
    from glob import glob

    fmri_paths = pd.read_csv(path.join(bids_dir, 'conversion_info', 'fmri_paths.tsv'), sep='\t')

    fmri_paths = fmri_paths[fmri_paths.Path.notnull()]

    fmri_paths['BIDS_SubjID'] = ['sub-ADNI' + s.replace('_', '') for s in fmri_paths.Subject_ID.to_list()]
    fmri_paths['BIDS_Session'] = ['ses-' + s.replace('bl', 'm00').upper() for s in fmri_paths.VISCODE.to_list()]

    count = 0
    count_wrong = 0
    name_wrong = 0
    no_dir = 0

    for r in fmri_paths.iterrows():
        image = r[1]
        image_dir = path.join(bids_dir, image.BIDS_SubjID, image.BIDS_Session, 'func')

        if not path.isdir(image_dir):
            no_dir += 1
            # continue

        image_pattern = path.join(image_dir, '%s_%s_*bold*' % (image.BIDS_SubjID, image.BIDS_Session))
        files_list = glob(image_pattern)

        if not files_list:
            # print("No images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            count += 1

        elif len(files_list) != 2:
            print("Wrong files count for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            print(files_list)
            count_wrong += 1
        elif sum([not f.endswith(('_task-rest_bold.json', '_task-rest_bold.nii.gz')) for f in files_list]) > 0:
            name_wrong += 1
            print(files_list)

    print(count)
    print(count_wrong)
    print(name_wrong)
    print(no_dir)
